
/***************************************************************************
    libBIOS - A portable library for developing an OpenSource BIOS
    Copyright (C) 2001 - David Aparicio

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author can be reached at 'daparic@terra.es'
****************************************************************************/
// 
#ifndef _EMOS_H
#define _EMOS_H

#include "libEMOS.h"

//------------------------ TRAPS
#define	TRAP_KNL_RESET	0	// Inicializacion del kernel
#define	TRAP_KNL_EXIT	1

#define	TRAP_MSG_NEW 	2
#define	TRAP_MSG_FREE 	3
#define	TRAP_MSG_JOIN 	4
#define	TRAP_MSG_SPLIT 	5

#define	TRAP_MSG_SEND	6
#define	TRAP_MSG_RECV	7

#define	TRAP_MSG_BIND	8
#define	TRAP_MSG_UNBIND	9

//------------------------
#define JIFFIE		1	// TIC frequency (miniOS), MSECS

// Scale factor for JIFFIE(1 ms) to TIMER CLOCK CYCLES
#define CYCLES(M)	(M*3125)

#ifdef __ASSEMBLER__
//---- Offset for REGS storage in context (ASSEMBLER)
#	define O_R0		 0
#	define O_R1		 4
#	define O_R2		 8
#	define O_R3		12
#	define O_R4		16
#	define O_R5		20
#	define O_R6		24
#	define O_R7		28
#	define O_R8		32
#	define O_R9		36
#	define O_R10		40
#	define O_R11		44
#	define O_R12		48
#	define O_R13		52	// user SP
#	define O_R14		56	// user LR
#	define O_FL		60	// user PSR
#	define O_PC		64	// user PC
#else
//---- Index for REGS storage in context (C LANGUAGE)
#	define I_PC   16	// Program counter [O_PC/4]
#	define I_SP   13	// Stack pointer   [O_R13/4]
#	define I_RT   0		// Return value    [O_R0 /4]
#	define I_A0   0		// Argument 0	   [O_R0 /4]
#endif

#define REGS_SIZE	18	// maximum storage capability of context
#define TCBQ_SIZE	128	// maximum storage capability of receiving queue
#define TCBS_SIZE	16	// maximum storage capability of signaling queue
#define STACK_SIZE	2048	// 1024

#define MAX_TCB		32	// maximo de TCB en el sistema
#define MAX_MCB		2048	// estimado reserva + circulantes
#define MAX_MUTEX	16
#define MAX_TIMER	(MAX_TCB*2+MAX_MCB)

#define TIC2MSEC(A)	(A)
#define MSEC2TIC(A)	((A)==TM_NOTIMEOUT ? TM_NOTIMEOUT : (A/JIFFIE))

// Offsets en estructura SYS_T (para ensamblador)
#define SYS_RUNNING	0
#define SYS_CALLING	4
#define SYS_TOTALTIC	8
#define SYS_TCB		12

#ifndef __ASSEMBLER__

typedef enum {
	STM_USER=0x40000001,	// Creado, responsabilidad de usuario
	STM_COMM=0x40000002,	// Encolado (por MSGcomm): NO CAMBIAR VALOR
	STM_SEND=0x40000003,	// Encolado (por MSGsend): NO CAMBIAR VALOR
	STM_SIGN=0x40000004,	// Encolado (por signal) : NO CAMBIAR VALOR
	STM_RESP=0x40000005,	// Encolado (por MSGresp)
	STM_FREE=0x40000000 	// En cola de libres
	} statM_t;

typedef struct {		// Conservar alineacion 512 bytes...
	//--- Datos "de usuario" (DEBE ESTAR LO PRIMERO Y EN ESE ORDEN!!!)
	msghdr_t	msghdr;		// Definido en "rtos.h"

	//--- Datos "de kernel"
	//--- Agrupaciones
	union _mcb_t	*p, *n;		// Enlaces para agrupaciones de MCB

	//--- Mensajes multipunto (solo para primeros 32 TCB)
	u32		mask;		// "1" para entregar en ese TCB
					// "0" cuando el mensaje se ha leido

	//--- Shortcut de temporizador (solo mensajes COMM)
	struct _timer_t *tmMcb;
	} mcbhdr_t;

#define MCBHDR(MCB) 	(MCB)->m.h.hdr
#define MCB_NELEM(MCB)		MCBHDR(MCB).msghdr.nelem
#define MCB_LENGTH(MCB)		MCBHDR(MCB).msghdr.length
#define MCB_STATUS(MCB)		MCBHDR(MCB).msghdr.status
#define MCB_FROM(MCB)		MCBHDR(MCB).msghdr.from
#define MCB_SLOT(MCB)		MCBHDR(MCB).msghdr.slot

#define MCB_MASK(MCB)	MCBHDR(MCB).mask
#define MCB_P(MCB)	MCBHDR(MCB).p
#define MCB_N(MCB)	MCBHDR(MCB).n

#define IS_MCB_NEW(MCB)		(MCB_STATUS(MCB)==STM_USER)
#define IS_MCB_COMM(MCB)	(MCB_STATUS(MCB)==STM_COMM)
#define IS_MCB_SEND(MCB)	(MCB_STATUS(MCB)==STM_SEND)
#define IS_MCB_RESP(MCB)	(MCB_STATUS(MCB)==STM_RESP)
#define IS_MCB_SIGN(MCB)	(MCB_STATUS(MCB)==STM_SIGN)

#define MCB2BUF(A)	(u8    *)(A ? A->m.d : NULL)
#define BUF2MCB(A)	(mcb_t *)(A ? (u8   *)(A) - HDR_ALIGN : NULL)

typedef union _mcb_t {
	// Para alinear esta estructura exactamente a 4 Kbytes
	u8	dummy[MCB_ALIGN];
	struct {
		// Campo de sistema (alineado a 512 bytes)
		union {
			u8		dummy[HDR_ALIGN];
			mcbhdr_t	hdr;
			} h;
		// Campo accesible por el usuario (son 2 Kbytes alineados a 512)
		u8      d[MAX_BLOCK];
		} m;
	} mcb_t;

typedef enum {
	STT_READY=0x30000001,	// Disponible para ejecutar
	STT_WAITS=0x30000002,	// En espera por COMM
	STT_WAITR=0x30000003,	// En espera por RECV
	STT_FREE =0x30000000	// Recurso libre
	} statT_t;

#define IS_EMPTYQ(Q)	((Q).get==(Q).put)

typedef struct {
	u32	get, put;
	mcb_t  *q[TCBQ_SIZE];
	} rq_t;
	
typedef struct _tcb {
	// *** Estos campos deben conservar el orden en la estructura
	u32		regs[REGS_SIZE];	// Contexto de CPU
				// regs[0] es RET-ADDR
				// regs[1] es STACK
				// regs[2] es RET-VALUE
	// *** A partir de aqui el orden no importa
	//--- Local stack for this task
	u32		stack[STACK_SIZE];

	//--- Rendezvous queues
	rq_t		send;		// Receiving point for streams
	rq_t		comm;		// Receiving point for messages
	rq_t		wait;		// Receiving point for signals
	u8		**msg;		// MSG hint for delayed-MSGrecv

	//--- Parte sincrona
	void    	(*poll)(u32 totaltic);
	u32		polltics;

	//--- Estado de la tarea
	statT_t		status;		// statT_t

	//--- Shortcuts a los temporizadores activos en este TCB
	struct _timer_t *tmAsyn;
	struct _timer_t *tmSync;

	//--- Lista doble enlazada
  	struct _tcb	*p, *n;
	} tcb_t;

// Enlaces para TIMER (MCB y TCB)

typedef struct _timer_t {
	struct _timer_t *p, *n;
	u32	tics_left;
#define	TIMER_SYNC	1	// Temporizador TCB sincrono (poll)
#define	TIMER_ASYN	2	// Temporizador TCB asincrono (loop)
#define	TIMER_MCB	3	// Temporizador MCB
	u32	tipo;
	union {
		mcb_t	*mcb;
		tcb_t	*tcb;
		} i;
	} timer_t;
	
typedef struct _mutex_t {
		tcb_t	*owner;
		tcb_t	*waitq;
	} mutex_t;

#define SET_BIT(M, N)	M |=  (1<<(N))
#define CLR_BIT(M, N)	M &= ~(1<<(N))
#define IS_BITS(M, N)	(M & (1<<(N)))

// No se pueden generar mas de 66 transferencias en media por tic...
// Supongamos un pico acumulable de 40 ms
// El maximo de operaciones simultaneas es por tanto menor que:
#define MAX_OPS (66*MSEC2TIC(40))

typedef struct {
	// Estos campos deben conservar el orden en la estructura
	tcb_t  *running;
	tcb_t  *calling;
	u32     tic;		// TotalTICS
	tcb_t	TCB[MAX_TCB];

	// A partir de aqui el orden no importa
	mcb_t	*MCB;
	timer_t	TMR[MAX_TIMER];

	tcb_t	*ready[MAX_PRIO];
	tcb_t	*Tfree;
	mcb_t	*Mfree;
	timer_t	*timer;	// temporizaciones de TCB o MCB
	timer_t	*Efree;

	// Reserva de memoria para recursos
	mutex_t	mutex [MAX_MUTEX];
	u32	channel[MAX_SLOT];
	} sys_t;

// *** Entrada en la tabla de tareas...
typedef struct _stask_t {
	void    (*init)(pid_t pid);
	u32	polltics;
	void    (*poll)(u32 totaltic);
	u32	prio;	// Entre 0..MAX_PRIO-1
	u32	cpu;
	} stask_t;

//--- tcb.c: Operaciones con TCB
void   tcbReady(tcb_t *tcb);
void   tcbWait(tcb_t *tcb, u8 **msg);

//--- mbc.c: Operaciones con MCB
mcb_t *mcbNew  (u32 nelem);
void   mcbRespFree(tcb_t *tcb, mcb_t *mcb);
void   mcbCommSend(tcb_t *tcb, mcb_t *mcb, u32 msecs);
mcb_t *mcbRecv (rq_t *rq);
void   mcbJoin (mcb_t **mcbD, mcb_t **mcbS);
void   mcbSplit(mcb_t **mcbD, mcb_t **mcbS);

//--- timer.c: Operaciones con TIMER
void   timerInsert(u32 tipo, void *xcb, u32 msecs);
void   timerDelete(timer_t *tm);

// --- Inicializacion de contexto TCB (multi.S)
void	initTimer(u32 milis_por_int);
void	initRegs(u32 *regs, pid_t pid, void *progaddr, u32 *stackaddr);

// Modulo swi.S
//void   *swiMSGnew (void);
//void    swiMSGfree(void *msg);
//void   *swiMSGrecv(u32 msecs);
//void    swiMSGsend(void *msg, u32 dest);

//--- kernel.c: syscalls & stubs

int     sysMSGnew  (u8 **msg, u32 nmsg);
int     sysMSGfree (u8  *msg);
int     sysMSGjoin (u8 **msgD, u8 **msgS);
int     sysMSGsplit(u8 **msgD, u8 **msgS);

int     sysMSGbind  (pid_t pid, u32 group);
int     sysMSGunbind(pid_t pid, u32 group);

int     sysMSGsend(u8  *msg, u32 group);
int     sysMSGrecv(u8 **msg, u32 msecs, u32 mask);

// Proceso ZERO (IDLE y poll interno del miniOS)
void   USR0sync(u32 tic);	// Poll DMA y FIFOS IPC (KERNEL)
void   USR0asyn(pid_t pid);	// Proceso IDLE (KERNEL)

// Stub para tratamiento IRQ-TIC
void   KNLirq(void);

//--- kernel.c: Enlaces con la crt0.S
// Stub para tratamiento SWI-TRAP
void   KNLswi(u32 a1, u32 a2, u32 a3, u32 trap);

//--- variables ---
extern sys_t sys;		// El alineado no es importante
extern const stask_t tasks[];

#endif	// ASSEMBLER
#endif	// _EMOS_H

