/***************************************************************************
    libBIOS - A portable library for developing an OpenSource BIOS
    Copyright (C) 2001 - David Aparicio

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author can be reached at 'daparic@terra.es'
****************************************************************************/
// 
#ifndef _LIBEMOS_H
#define _LIBEMOS_H

// Higher number is better priority
#define MAX_PRIO	4

// Special timer constants
#define TM_NOWAIT	0x00000000
#define TM_NOTIMEOUT	0xFFFFFFFF

// Process ID constants
#define PID_NULL	0xFFFFFFFF

// Memory alignments (MANDATORY in order DMA to work)
#define	MCB_ALIGN	4096
#define	HDR_ALIGN	512

// Max user block for MCB passing
#define	MAX_BLOCK	2048

// Max of tasks in emOS.
#define MAX_PID		32		// Fixed for a 32bit processor

// Max of streams in emOS
#define MAX_SLOT	(MAX_PID+64)	// maximum concurrent stream capability
					// -> 0..MAX_PID-1 are "single point"

// Syscall results
#define R_OK	(int) 0		// Everything is fine
#define R_EINTR	 (int)-1	// Caller will be frozen, no inmediate result
#define R_EAGAIN (int)-2	// There is no result - timeout
#define R_ENOMEM (int)-3	// There is no resources
#define R_EFAULT (int)-4	// Bad parameters

#include "libBIOS.h"

#ifndef __ASSEMBLER__

typedef u32 pid_t;

//---------------------- PRIVATE: NOT FOR USER -------------------
typedef struct {
	u32		length;		// Size of user data
	u32		nelem;		// Number of "brother" MCBs in this grp
	pid_t		from,		// 0..MAX_PID-1
			slot;		// 0..MAX_SLOT-1
	u32		status;		// same as status
	} msghdr_t;

// Macros for header access
#define _MSGHDR(BUF)	((msghdr_t *)(((u8 *)(BUF)) -  HDR_ALIGN))

//---------------------- PUBLIC: USER INTERFACE -------------------
#define MSG_LENGTH(BUF)		_MSGHDR(BUF)->length
#define MSG_NELEM(BUF)		_MSGHDR(BUF)->nelem
#define MSG_FROM(BUF)		_MSGHDR(BUF)->from

#define	IS_CONTROL(BUF)	(_MSGHDR(BUF)->status == 0x40000002)
#define	IS_STREAM(BUF)	(_MSGHDR(BUF)->status == 0x40000003)
#define	IS_SIGNAL(BUF)	(_MSGHDR(BUF)->status == 0x40000004)

//-- Receiver masking (mask in MSGrecv)
#define FL_NONE		0
#define FL_CONTROL	1
#define FL_STREAM	2
#define FL_SIGNAL	4
#define FL_ANY		(FL_CONTROL|FL_STREAM|FL_SIGNAL)

extern u32 _errno;

//---------------------- PRIVATE: NOT FOR USER -------------------
//-- Block (MCB) management
int   _emOSnew  (u8 **msg, u32 nelem);
int   _emOSfree (u8 *msg);
int   _emOSjoin (u8 **msgD, u8 **msgS);
int   _emOSsplit(u8 **msgD, u8 **msgS);

//-- Multicast messaging
int   _emOSbind  (pid_t pid, u32 channel);
int   _emOSunbind(pid_t pid, u32 channel);

//-- Unicast/multicast messaging
int   _emOSsend(u8  *msg, u32 msecs);
int   _emOSrecv(u8 **msg, u32 msecs, u32 mask);

//---------------------- PUBLIC: USER INTERFACE -------------------
//--- Helpers for syscalls: They avoid using stack for passing parms
INLINE	u8 *NEW(u32 nelem)
	{
		u8 *tmp = NULL;
		_errno = _emOSnew(&tmp, nelem);
		return _errno==R_OK ? tmp : NULL;
	}

INLINE	int FREE(u8 *msg)	// Counterpart for STREAM/NEW
	{
		_errno = _emOSfree(msg);
		return _errno;
	}

INLINE  int JOIN(u8 **dest, u8 **orig)
	{
		_errno = _emOSjoin(dest, orig);
		return _errno;
	}

INLINE  int SPLIT(u8 **dest, u8 **orig)
	{
		_errno = _emOSsplit(dest, orig);
		return _errno;
	}

INLINE  int BIND(u32 suscriber, u32 channel)
	{
		_errno = _emOSbind(suscriber, channel);
		return _errno;
	}

INLINE  int UNBIND(u32 suscriber, u32 channel)
	{
		_errno = _emOSunbind(suscriber, channel);
		return _errno;
	}

INLINE  u8 *RECEIVE(u32 msecs, u32 mask)
	{
		u8 *tmp = NULL;
		_errno = _emOSrecv(&tmp, msecs, mask);
		return _errno==R_OK ? tmp : NULL;
	}

INLINE	int CONTROL(u8 *msg, u32 channel, u32 length, u32 msecs)
	{
		_MSGHDR(msg)->slot   = channel;
		_MSGHDR(msg)->length = length;
		_errno = _emOSsend(msg, msecs);
		return _errno;
	}

INLINE	int STREAM (u8 *msg, u32 channel, u32 length)
	{
		_MSGHDR(msg)->slot   = channel;
		_MSGHDR(msg)->length = length;
		_errno = _emOSsend(msg, 0);
		return _errno;
	}

INLINE	int SLEEP  (u32 msecs)
	{
		_errno = _emOSrecv(NULL, msecs, FL_NONE);
		return _errno;
	}

INLINE	int SIGNAL (u8 *msg, u32 length)
	{
		_MSGHDR(msg)->slot   = PID_NULL;
		_MSGHDR(msg)->length = length;
		_errno = _emOSsend(msg, 1);
		return _errno;
	}

INLINE	int REPLY  (u8 *msg, u32 length)	// Counterpart for CONTROL
	{
		_MSGHDR(msg)->length = length;
		_errno = _emOSfree(msg);
		return _errno;
	}

INLINE	int RELEASE(u8 *msg)	// Counterpart for SIGNAL (same as FREE)
	{
		_errno = _emOSfree(msg);
		return _errno;
	}

#endif	// __ASSEMBLER__
#endif	// _LIBEMOS_H
