/***************************************************************************
    libBIOS - A portable library for developing an OpenSource BIOS
    Copyright (C) 2001 - David Aparicio

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author can be reached at 'daparic@terra.es'
****************************************************************************/
// 
#include "emOS.h"

#undef DEPURACION
#undef DEPURACION_S

extern u32 _user_mode;

#ifdef DEPURACION
const char *_strmsec(u32 msecs)
{
	static char bf[128];

	if(msecs==TM_NOTIMEOUT)
		return "ETERNAL";
	else
		sprintf(bf, "%i", msecs);
	return bf;
}
#endif

//-------------------------------------------------------

int sysMSGnew (u8 **msg, u32 nmsg)
{
	mcb_t *mcb;

#ifdef DEPURACION
	printk("-<-%i.tcb(%i).sys%s(NEW, %i)\n",
		sys.tic, sys.calling - sys.TCB,
		_user_mode ? "Trap" : "Call", nmsg);
#endif
	if (*msg || !nmsg)
	{
#ifdef DEPURACION
		printk("->-EFAULT\n");
#endif
		return R_EFAULT;
	}
	mcb = mcbNew(nmsg);
	MCB_FROM(mcb)   = sys.calling - sys.TCB;
	MCB_SLOT(mcb)   = PID_NULL;
	MCB_LENGTH(mcb) = MAX_BLOCK;
	*msg = MCB2BUF(mcb);
#ifdef DEPURACION
	printk("->-OK mcb(%i)\n", mcb-sys.MCB);
#endif
	return R_OK;
}

int sysMSGjoin(u8 **msgD, u8 **msgS)
{
	mcb_t *mcbS = BUF2MCB(*msgS);
	mcb_t *mcbD = BUF2MCB(*msgD);

	if(!*msgS || !*msgD)
		return R_EFAULT;
	mcbJoin(&mcbD, &mcbS);
	*msgS = MCB2BUF(mcbS);
	*msgD = MCB2BUF(mcbD);
	return R_OK;
}

int sysMSGsplit(u8 **msgD, u8 **msgS)
{
	mcb_t *mcbS = BUF2MCB(*msgS);
	mcb_t *mcbD = BUF2MCB(*msgD);

	if(!*msgS)
		return R_EFAULT;
	mcbSplit(&mcbD, &mcbS);
	*msgS = MCB2BUF(mcbS);
	*msgD = MCB2BUF(mcbD);
	return R_OK;
}

int   sysMSGbind(pid_t pid, u32 group)
{
	if(pid >= MAX_PID || group<MAX_PID || group >= MAX_SLOT)
		return R_EFAULT;
	SET_BIT(sys.channel[group], pid);
	return R_OK;
}

int   sysMSGunbind(pid_t pid, u32 group)
{
	if(pid >= MAX_PID || group<MAX_PID || group >= MAX_SLOT)
		return R_EFAULT;
	CLR_BIT(sys.channel[group], pid);
	return R_OK;
}

int   sysMSGsend(u8 *msg, u32 msecs)
{
	u32    status;
	mcb_t *mcb = BUF2MCB(msg);

	// Auto-sent message
	if(MCB_SLOT(mcb)==PID_NULL)
		MCB_SLOT(mcb) = sys.calling - sys.TCB;
#ifdef DEPURACION
	printk("-<-%i.tcb(%i).sys%s(SEND, mcb(%i), %s)\n",
		sys.tic, sys.calling - sys.TCB, _user_mode ? "Trap" : "Call",
		mcb - sys.MCB,
		_strmsec(msecs));
#endif
/*------------------------------------
	if msecs==0       --> STREAM
	else if user_mode --> MESSAGE
	else              --> SIGNAL
  ------------------------------------
*/
	// Free undeliverable messages (no receivers listening)
	if(!(MCB_MASK(mcb)  = sys.channel[MCB_SLOT(mcb)]))
	{
		mcbRespFree(NULL, mcb);
#ifdef DEPURACION
		printk("->-OK\n");
#endif
		return R_OK;
	}
	
	if(!msecs)	// STREAM mode
	{
		u32 i, m;

	// Reject null messages and pending-to-deliver messages
		if(!msg || (MCB_STATUS(mcb)!=STM_USER
				&& MCB_STATUS(mcb)!=STM_SIGN))
		{
#ifdef DEPURACION
			printk("->-EFAULT\n");
#endif
			return R_EFAULT;
		}
	// As an exception, a signalled message *can* be "SENT"

		status = STM_SEND;
		for(i=0, m=1; i<MAX_PID; i++, m<<=1)	// STM_SEND
			if(MCB_MASK(mcb) & m)
				mcbCommSend(sys.TCB + i, mcb, 0);
	}
	else		// MESSAGE or SIGNAL mode
	{
		// Reject null messages and pending-to-deliver messages
		// Allow only unicasted messages
		if(!msg || MCB_STATUS(mcb)!=STM_USER || MCB_SLOT(mcb)>=MAX_PID)
		{
#ifdef DEPURACION
			printk("->-EFAULT\n");
#endif
			return R_EFAULT;
		}
		mcbCommSend(sys.TCB + MCB_SLOT(mcb), mcb, msecs);
		if(_user_mode)		// It's a COMMAND, set TIMER
		{
			timerInsert(TIMER_MCB, mcb, msecs);
			tcbWait(sys.calling, NULL);
		}
	}
#ifdef DEPURACION
	if(sys.calling==sys.running)
		printk("->-OK\n");
	else
		printk("->-EINTR switch(%i->%i)\n",
			sys.calling-sys.TCB, sys.running-sys.TCB);
#endif
	return (sys.calling==sys.running) ? R_OK : R_EINTR;
}

int  sysMSGfree(u8 *msg)
{
	mcb_t *mcb = BUF2MCB(msg);

#ifdef DEPURACION
	printk("-<-%i.tcb(%i).sys%s(FREE, mcb(%i))\n",
		sys.tic, sys.calling - sys.TCB, _user_mode ? "Trap" : "Call",
		mcb - sys.MCB);
#endif
	/*--- Si el mensaje es:
		 NEW: Simplemente liberarlo
		SEND: Marcar como leido. Si no quedan marcas, liberarlo.
		COMM: Simplemente liberarlo
		RESP: Devolverlo al enviante (si WAITS) o liberarlo.
	*/
	if (!msg)
	{
#ifdef DEPURACION
		printk("->-EFAULT\n");
#endif
		return R_EFAULT;
	}
	if(IS_MCB_COMM(mcb))	// RESP to MESSAGE
	{
		timerDelete(MCBHDR(mcb).tmMcb);
		mcbRespFree(sys.TCB + MCB_FROM(mcb), mcb);
#ifdef DEPURACION
		printk("->-OK\n");
#endif
		return R_OK;
	}
	if(IS_MCB_SIGN(mcb))	// RELEASE the MESSAGE
	{
		MCB_STATUS(mcb) = STM_USER;
#ifdef DEPURACION
		printk("->-OK\n");
#endif
		return R_OK;
	}
	if(IS_MCB_SEND(mcb))	// if STREAM, return if we are not the last
	{
		u32 idx = sys.calling - sys.TCB;
		if(idx<MAX_PID)
		{
			MCB_MASK(mcb) &= ~(1<<idx);
			if(MCB_MASK(mcb))
			{
#ifdef DEPURACION
				printk("->-OK\n");
#endif
				return R_OK;
			}
		}
	}
	//-- Free the message (FREE in RESP or last SEND)
	mcbRespFree(NULL, mcb);
#ifdef DEPURACION
	printk("->-OK\n");
#endif
	return R_OK;
}

int sysMSGrecv(u8 **msg, u32 msecs, u32 mask)
{
	mcb_t *mcb;

#ifdef DEPURACION
	printk("-<-%i.tcb(%i).sys%s(RECV, %s, %cSGN%cCTL%cDAT)\n",
		sys.tic, sys.calling - sys.TCB, _user_mode ? "Trap" : "Call",
		_strmsec(msecs),
		(mask&FL_SIGNAL ) ? '+' : '-',
		(mask&FL_CONTROL) ? '+' : '-',
		(mask&FL_STREAM ) ? '+' : '-');
#endif
	//--- If there are queued messages, then receiving priority is:
	//		* SIGNALS, then
	//		* CONTROL, then
	//		* STREAMS
	if((mask&FL_SIGNAL) && (mcb = mcbRecv(&sys.calling->wait)))
	{
		*msg = MCB2BUF(mcb);
#ifdef DEPURACION
		printk("->-OK signaled.mcb(%i)\n", mcb - sys.MCB);
#endif
		return R_OK;
	}
	if((mask&FL_CONTROL) && (mcb = mcbRecv(&sys.calling->comm)))
	{
		*msg = MCB2BUF(mcb);
#ifdef DEPURACION
		printk("->-OK control.mcb(%i)\n", mcb - sys.MCB);
#endif
		return R_OK;
	}
	if((mask&FL_STREAM) && (mcb = mcbRecv(&sys.calling->send)))
	{
		*msg = MCB2BUF(mcb);
#ifdef DEPURACION
		printk("->-OK streamed.mcb(%i)\n", mcb - sys.MCB);
#endif
		return R_OK;
	}
	if(!msecs)
	{
		*msg = NULL;
#ifdef DEPURACION
		printk("->-EAGAIN\n");
#endif
		return R_EAGAIN;
	}

	//--- No queued messages, sleep
	timerInsert(TIMER_ASYN, sys.calling, msecs);
	tcbWait(sys.calling, msg);
#ifdef DEPURACION
	printk("->-EINTR switch(%i->%i)\n",
		sys.calling-sys.TCB, sys.running-sys.TCB);
#endif
	return R_EINTR;
}

//-------------------------------------------------------

void	USR0asyn(pid_t mypid) { while(1) {} }

//-------------------------------------------------------

INLINE  void SYSpoll(void)
{
	register mcb_t *mcb;
	register tcb_t *tcb;
	void *savecaller = sys.running;
#ifdef DEPURACION
	u32 hasprinted = 0;
#endif

	sys.tic++;
	_user_mode = 0;
        if (sys.timer)
	{
		sys.timer->tics_left--;
#ifdef DEPURACION
		if(!sys.timer->tics_left)
#ifndef DEPURACION_S
			if(sys.timer->tipo!=TIMER_SYNC)
#endif
			{
				hasprinted = 1;
				printk("-<-%i.sysTrap(TICK)\n", sys.tic);
			}
#endif
	}
	while(sys.timer)
	{
		register timer_t *tm = sys.timer;

		//--- Dejar de examinar cuando ya no haya mas vencidos
		if(tm->tics_left)
			break;

		//--- Temporizador vencido
		timerDelete(tm);
		switch(tm->tipo)
		{
		case TIMER_ASYN:	// TCB asincrono
			tcb = tm->i.tcb;
			tcbReady(tcb);	// Despertar
			tcb->msg    = NULL;
			tcb->regs[I_RT] = R_EAGAIN;
			break;
		case TIMER_SYNC:	// TCB sincrono (call poll)
			tcb = tm->i.tcb;
			sys.calling = tcb;
			tcb->poll(sys.tic);
			timerInsert(TIMER_SYNC, tcb, tcb->polltics);
			break;
		case TIMER_MCB:		// MCB vencido (MSGcomm)
			mcb = tm->i.mcb;
			tcb = sys.TCB + MCB_FROM(mcb);
			mcbRespFree(tcb->status==STT_WAITS ? tcb : NULL, mcb);
			break;
		}
	}	// while(tm)
	_user_mode = 1;
	sys.calling = savecaller;

	// Seleccion RR de planificador
        if(sys.calling==sys.running && sys.running != sys.running->n)
        {
                sys.running = sys.running->n;
        }
#ifdef DEPURACION
	if(hasprinted)
		printk("->-OK\n");
#endif
}

//-------------------------------------------------------
u32	_user_mode;
sys_t	sys;

#define ROUNDUP(A,B)	(void *)(((u32)A + B) & ~(B-1))

static void KNLinit(env_t *env)
{
	int i;
	u32 npoll, nloop;

// user_mode: Para que las "syscall" valgan para parte sincrona y asincrona...
// Vale 0 cuando el usuario esta en modo supervisor (sincrono)
// Vale 1 cuando el usuario esta en modo usuario    (asincrono)
	_user_mode = 1;

	sys.MCB = (void *)ADDR_MCB;
	// ----------- TCB related
	for(i=0; i<MAX_TCB;  i++)
	{
		tcb_t *p  = sys.TCB + i;
		p->n = (i==MAX_TCB-1) ? sys.TCB           : p+1;
		p->p = (i==0)         ? sys.TCB+MAX_TCB-1 : p-1;
		p->status = STT_FREE;
	}
	for(i=0; i<MAX_MUTEX; i++)
	{
		sys.mutex[i].owner = NULL;
		sys.mutex[i].waitq = NULL;
	}
	for(i=0; i<MAX_PRIO ; i++) sys.ready[i]       = NULL;
	sys.Tfree   = sys.TCB; // Todas los recursos en 'zombie'

	// ----------- MCB related
	for(i=0; i<MAX_MCB;  i++)
	{
		mcb_t *p = sys.MCB + i;

		MCB_N(p)     = (i==MAX_MCB-1) ? sys.MCB           : p+1;
		MCB_P(p)     = (i==0)         ? sys.MCB+MAX_MCB-1 : p-1;
		MCB_STATUS(p)= STM_FREE;
	}
	sys.Mfree  = sys.MCB;

	// ----------- TIMER related
	for(i=0; i<MAX_TIMER;  i++)
	{
		timer_t *p = sys.TMR + i;
		p->n = (i==MAX_TIMER-1) ? sys.TMR             : p+1;
		p->p = (i==0)           ? sys.TMR+MAX_TIMER-1 : p-1;
		p->tipo = 0;
	}
	sys.Efree  = sys.TMR;
	sys.timer  = NULL;

	// ----------- MULTIPOINT related
	for(i=0; i<MAX_SLOT;  i++)
		sys.channel[i] = i<MAX_PID ? (1<<i) : 0;

	// Initialize "running" to idle task
	sys.running = sys.TCB;
	// TIC inicialization. Goes here because of pretty printing
	// in "tcbReady" (ifdef DEPURACION)
	sys.tic = 0;

	// Inicializacion de tareas
	for(i=0, nloop=0, npoll=0; tasks[i].init; i++)
	{
		tcb_t *tcb = sys.TCB + i;

		initRegs(tcb->regs, i, tasks[i].init, tcb->stack);
		tcb->poll       = tasks[i].poll;
		tcb->polltics   = tasks[i].polltics;
		if(tcb->poll)
		{
			timerInsert(TIMER_SYNC, tcb, tcb->polltics);
			npoll++;
		}
		nloop++;
		tcbReady(tcb);
	}

	// First context switch. We put an impossible value on "calling"
	// to force it
	sys.calling = sys.TCB + nloop;

	// Re-initialize malloc/free library
	// (LOADER already inicialized in another region)
	IniHeap(ADDR_TEMP, SZ_8M);

	printk("=(3)========================================\n");
	printk("            [FW"DATE" MINIOS]\n");
	printk("============================================\n");
	printk("CORE: %i proc(s), %i poll(s)\n", nloop, npoll);
	initTimer(CYCLES(JIFFIE));
}

#ifdef DEPURACION_S
const char strcmd[][64] =
		{	"INIT",  "HALT",
			"NEW",  "FREE",  "JOIN", "SPLIT",
			"SEND", "RECV",  "BIND", "UBIND"
		};
#endif

void KNLirq(void)
{
	SYSpoll();
#ifdef DEPURACION_S
	// --- depuracion ---
	if(sys.calling != sys.running)
	{
		printk("%9i-%i] %5s (prev %i)\n\n",
			sys.tic, sys.running-sys.TCB,
			 "TIC", sys.calling-sys.TCB);
	}
#endif
}

void KNLswi(u32 a1, u32 a2, u32 a3, u32 num)
{
	register int r;
#ifdef DEPURACION_S
	printk("[[%i %08X(%i)->%08X(%i)\n", sys.TCB[1].status,
		sys.calling, sys.calling-sys.TCB,
		sys.running, sys.running-sys.TCB);
#endif
	r = R_OK;
	sys.calling = sys.running;
	switch(num)
	{
	case TRAP_KNL_RESET: KNLinit((env_t *)a1); break;
	case TRAP_KNL_EXIT:  exit(0);	           break;

	case TRAP_MSG_NEW :  r = sysMSGnew  ((u8 **)a1, a2);        break;
	case TRAP_MSG_FREE:  r = sysMSGfree ((u8  *)a1);            break;
	case TRAP_MSG_JOIN:  r = sysMSGjoin ((u8 **)a1, (u8 **)a2); break;
	case TRAP_MSG_SPLIT: r = sysMSGsplit((u8 **)a1, (u8 **)a2); break;

	case TRAP_MSG_SEND:  r = sysMSGsend ((u8  *)a1, a2);     break;
	case TRAP_MSG_RECV:  r = sysMSGrecv ((u8 **)a1, a2, a3); break;

	case TRAP_MSG_BIND:  r = sysMSGbind  (a1, a2);       break;
	case TRAP_MSG_UNBIND:r = sysMSGunbind(a1, a2);       break;

	default:
			CPUdump(sys.calling, num, a1);
			sys.running = sys.running->n;
			break;
	}
	if(sys.calling != sys.running)
		// Caller has been frozen, initialize RETVALUE
		sys.calling->regs[I_RT] = R_OK;
	else
		// Caller is still running, give him RETVALUE
		sys.calling->regs[I_RT] = r;

#ifdef DEPURACION_S
	printk("%9i-%i] %5s\n\n",
		sys.tic, sys.running-sys.TCB,
		 strcmd[num]);
	printk("]]%i %08X(%i)->%08X(%i)\n\n", sys.TCB[1].status,
		sys.calling, sys.calling-sys.TCB,
		sys.running, sys.running-sys.TCB);
#endif
}

