
/***************************************************************************
    libBIOS - A portable library for developing an OpenSource BIOS
    Copyright (C) 2001 - David Aparicio

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author can be reached at 'daparic@terra.es'
****************************************************************************/
// 
#include "emOS.h"

#undef DEPURACION

//--- MERGE de un grupo de elementos en otro
INLINE void _merge(mcb_t **mcbD, mcb_t **mcbO)
{
	if(*mcbD)	// Hay elementos en destino -> MERGE
	{
		mcb_t *tmp;

		MCB_N(MCB_P(*mcbO)) = *mcbD;
		MCB_N(MCB_P(*mcbD)) = *mcbO;
		tmp = MCB_P(*mcbO);		// En tmp el PRIMERO de "mcb"
		MCB_P(*mcbO) = MCB_P(*mcbD);
		MCB_P(*mcbD) = tmp;
		MCB_STATUS(*mcbO) = MCB_STATUS(*mcbD);
	}
	else
		*mcbD = *mcbO;
	*mcbO = NULL;
}

//--- Pasar "nelem"; inclusive el activo y posteriores, de mcbO a mcbD.
//--- Se retorna el nuevo valor de mcbO. mcbD vale el valor antiguo de mcbO
INLINE int _split(mcb_t **mcbD, mcb_t **mcbO, u32 nelem)
{
	register mcb_t *p0, *p1;
	mcb_t *aux;

	if(!nelem) return 0;

	p0 = *mcbO;
	p1 =  p0;
	while(nelem)
	{
		nelem--;
		p1 = MCB_N(p1);
		if(p1 == p0)
			break;	// No hay recursos!!!
	}

	if(nelem)
		return -1;	// No toco nada. No hay recursos suficientes

	if(p0 != p1)
	{
		//--- p0 es PRIMERO de lista de "separados"  (a hacer merge)
		//--- p1 es PRIMERO de lista de "remanentes" (nuevo mcbO)
		aux = MCB_P(p1);
		MCB_N(MCB_P(p1)) = p0;
		MCB_P(p1)        = MCB_P(p0);
		MCB_N(MCB_P(p0)) = p1;
		MCB_P(p0)        = aux;
		//--- reorganizar punteros
		MCB_STATUS(p1) = MCB_STATUS(*mcbO);
		*mcbO = p1;
	}
	else
	{
		//--- Se mueven exactamente todos los elementos de mcbO
		//--- Los elementos a mover quedan en p0
		*mcbO = NULL;
	}

	if(*mcbD)
	{
		MCB_STATUS(p0) = MCB_STATUS(*mcbD);
		aux = p0;
		_merge(mcbD, &aux);
		p0 = aux;
	}

	*mcbD = p0;
	return 0;
}

INLINE mcb_t *_addQ(rq_t *rq, mcb_t *mcb)
{
	u32 i = (rq->put+1) % TCBQ_SIZE;

	if(i == rq->get)	// FIFO llena, el mensaje se pierde
	{
//printk("addQ: full queue\n");
		return NULL;
	}
	rq->q[rq->put] = mcb;
	rq->put = i;
//printk("addQ get=%i put=%i\n", rq->get, rq->put);
	return mcb;
}

INLINE mcb_t *_delQ(rq_t *rq)
{
	mcb_t *aux = rq->q[rq->get];

	if(rq->put == rq->get)
	{
//printk("delQ: empty queue\n");
		return NULL;
	}
	rq->get = (rq->get+1) % TCBQ_SIZE;
//printk("delQ: get=%i put=%i\n", rq->get, rq->put);
	return aux;
}

#ifdef DEPURACION
const char *_mlab[] = {
	"Pool",
	"User",
	"Comm",
	"Send",
	"Sign",
	"Resp"
	};
#endif

mcb_t *mcbNew(u32 nelem)
{
	mcb_t *mcb = NULL;

	if(_split(&mcb, &sys.Mfree, nelem) != 0) // Recursos agotados
            return NULL; 
#ifdef DEPURACION
	printk("   %i.tcb(%i).mcb(%i) %s -> User\n",
		sys.tic, sys.calling-sys.TCB,
		mcb-sys.MCB, _mlab[MCB_STATUS(mcb)&7]);
#endif
	MCB_STATUS(mcb) = STM_USER;
	return mcb;
}

void   mcbJoin  (mcb_t **mcbD, mcb_t **mcbS)
{
	if(!mcbS || !mcbD || !*mcbS) return;
	if(*mcbD)
		MCB_STATUS(*mcbS) = MCB_STATUS(*mcbD);
	_merge(mcbD, mcbS);
}

void   mcbSplit (mcb_t **mcbD, mcb_t **mcbS)
{
	mcb_t *mcb = NULL;

	if(!mcbS || !mcbD || !*mcbS) return;
	if(*mcbD)
		MCB_STATUS(*mcbD) = MCB_STATUS(*mcbS);
	_split(mcbD, mcbS, 1);	// Mueve un elemento de D a S
}

void  mcbRespFree(tcb_t *tcb, mcb_t *mcb)
{
#ifdef DEPURACION
	u32 old = MCB_STATUS(mcb);
#endif
	if(tcb && tcb->status==STT_WAITS)
	{
#ifdef DEPURACION
		printk("   %i.tcb(%i).mcb(%i) %s -> Resp\n",
			sys.tic, sys.calling-sys.TCB,
			mcb-sys.MCB, _mlab[old&7]);
#endif
		tcbReady(tcb);
		*(tcb->msg) = MCB2BUF(mcb);
		tcb->regs[I_RT] = R_OK;
		MCB_STATUS(mcb) = STM_RESP;
	}
	else
	{
#ifdef DEPURACION
		printk("   %i.tcb(%i).mcb(%i) %s -> Pool\n",
			sys.tic, sys.calling-sys.TCB,
			mcb-sys.MCB, _mlab[old&7]);
#endif
		MCB_STATUS(mcb) = STM_FREE;
		_merge(&sys.Mfree, &mcb);
	}
}

void  mcbCommSend(tcb_t *tcb, mcb_t *mcb, u32 msecs)
{
	extern u32 _user_mode;
#ifdef DEPURACION
	u32 old = MCB_STATUS(mcb);
#endif

	//--- If receiver is slept, give the MCB (don't queue)
	if(tcb->status == STT_WAITR)
	{
		tcbReady(tcb);
		*(tcb->msg) = MCB2BUF(mcb);
		tcb->regs[I_RT] = R_OK;
		timerDelete(tcb->tmAsyn);
	}
	else			// queue mode, recvr is not listening
		tcb->msg = NULL;
	MCB_STATUS(mcb) = !msecs
				? STM_SEND
				: _user_mode
					? STM_COMM
					: STM_SIGN;
#ifdef DEPURACION
	printk("   %i.tcb(%i).mcb(%i) %s -> %s\n",
		sys.tic, sys.calling - sys.TCB,
		mcb-sys.MCB, _mlab[old&7], _mlab[MCB_STATUS(mcb)]);
#endif
	if(!tcb->msg)
		switch(MCB_STATUS(mcb))
		{
		case STM_SEND:
			if(_addQ(&tcb->send, mcb)) return;
			break;
		case STM_COMM:
			if(_addQ(&tcb->comm, mcb)) return;
			break;
		case STM_SIGN:
			if(_addQ(&tcb->wait, mcb)) return;
			break;
		default:
			break;
		}
	else		// msg NOT NULL when mode "not queued"
		return;
	printk("mcbCommSend: Overrun\n");
	while(1) {}
}

mcb_t *mcbRecv(rq_t *rq)
{
	mcb_t *mcb;

	if(!(mcb = _delQ(rq))) return NULL;
#ifdef DEPURACION
	printk("   %i.tcb(%i).mcb(%i) Recv (%s)\n",
		sys.tic, sys.calling-sys.TCB,
		mcb-sys.MCB, _mlab[MCB_STATUS(mcb)&7]);
#endif
	return mcb;
}

