//----------------------------------------------------------------------------;
// GENERIC ARM FUNCTIONS - divide and remainder.                              ;
//                                                                            ;
// NOTES                                                                      ;
//                                                                            ;
// 1/ You may wish to make these functions part of your O/S kernel, replacing ;
//    the implementations here by branches to the relevant entry addresses.   ;
//                                                                            ;
// 2/ Each divide function is a div-rem function, returning the quotient in   ;
//    r0 and the remainder in r1. Thus (r0, r1) -> (r0/r1, r0%r1). This is    ;
//    understood by the C compiler.                                           ;
//                                                                            ;
// 3/ Because of its importance in many applications, divide by 10 is treated ;
//    as a special case. The C compiler recognises divide by 10 and generates ;
//    calls to __rt_{u,s}div10, as appropriate.                               ;
//                                                                            ;
// 4/ In each of these implementations we make a careful compromise between   ;
//    fastest and smallest. In some cases (e.g. divide), further unrolling of ;
//    loops will yield a speed-up. However, this affects cache and working-   ;
//    set locality and can result in a slow-down for some applications on     ;
//    some hardware. You are advised not to unroll further before considering ;
//    carefully your application and its interaction with your hardware.      ;
//    (e.g. ARM-2 vs ARM-3, external cache vs no external cache, fraction of  ;
//     run-time spent executing divide, etc.).                                ;
//                                                                            ;
//----------------------------------------------------------------------------;

	.globl	__divsi3
__divsi3:
//
// Signed divide of a2 by a1: returns quotient in a1, remainder in a2
// Quotient is truncated (rounded towards zero).
// Sign of remainder = sign of dividend.
// Also destroys a3, a4 and ip.
// Makes dividend and divisor non-negative, then does an unsigned divide;
// the proper signs get sorted out again at the end.
// The code is mostly as for udiv, except that the justification part is
// slightly simplified by knowledge that the dividend is in the range
// [0..#x80000000]. One register may be gained thereby.
	MOV	ip, a2
	MOV	a2, a1
	MOV	a1, ip

        MOVS    ip, a1
        BEQ     dividebyzero
        RSBMI   a1, a1, #0              // absolute value of divisor
        EOR     ip, ip, a2
        ANDS    a4, a2, #0x80000000
        ORR     ip, a4, ip, LSR #1
        // ip bit 31  sign of dividend (= sign of remainder)
        //    bit 30  sign of dividend EOR sign of divisor (= sign of quotient)
        RSBNE   a2, a2, #0              // absolute value of dividend

        MOV     a3, a1
        MOV     a4, #0
s_loop:
        CMP     a2, a3, ASL #0
        BLS     s_shifted0mod8
        CMP     a2, a3, ASL #1
        BLS     s_shifted1mod8
        CMP     a2, a3, ASL #2
        BLS     s_shifted2mod8
        CMP     a2, a3, ASL #3
        BLS     s_shifted3mod8
        CMP     a2, a3, ASL #4
        BLS     s_shifted4mod8
        CMP     a2, a3, ASL #5
        BLS     s_shifted5mod8
        CMP     a2, a3, ASL #6
        BLS     s_shifted6mod8
        CMP     a2, a3, ASL #7
        MOVHI   a3, a3, ASL #8
        BHI     s_loop
s_loop2:
        CMP     a2, a3, ASL #7
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #7
        CMP     a2, a3, ASL #6
s_shifted6mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #6
        CMP     a2, a3, ASL #5
s_shifted5mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #5
        CMP     a2, a3, ASL #4
s_shifted4mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #4
        CMP     a2, a3, ASL #3
s_shifted3mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #3
        CMP     a2, a3, ASL #2
s_shifted2mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #2
        CMP     a2, a3, ASL #1
s_shifted1mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #1
        CMP     a2, a3, ASL #0
s_shifted0mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #0
        CMP     a1, a3, LSR #1
        MOVLS   a3, a3, LSR #8
        BLS     s_loop2
        MOV     a1, a4
        TST     ip, #0x40000000
        RSBNE   a1, a1, #0
        TST     ip, #0x80000000
        RSBNE   a2, a2, #0
        MOV	pc, lr

	.globl	__modsi3
__modsi3:
//
// Signed divide of a2 by a1: returns quotient in a1, remainder in a2
// Quotient is truncated (rounded towards zero).
// Sign of remainder = sign of dividend.
// Also destroys a3, a4 and ip.
// Makes dividend and divisor non-negative, then does an unsigned divide;
// the proper signs get sorted out again at the end.
// The code is mostly as for udiv, except that the justification part is
// slightly simplified by knowledge that the dividend is in the range
// [0..#x80000000]. One register may be gained thereby.
	MOV	ip, a2
	MOV	a2, a1
	MOV	a1, ip

        MOVS    ip, a1
        BEQ     dividebyzero
        RSBMI   a1, a1, #0              // absolute value of divisor
        EOR     ip, ip, a2
        ANDS    a4, a2, #0x80000000
        ORR     ip, a4, ip, LSR #1
        // ip bit 31  sign of dividend (= sign of remainder)
        //    bit 30  sign of dividend EOR sign of divisor (= sign of quotient)
        RSBNE   a2, a2, #0              // absolute value of dividend

        MOV     a3, a1
        MOV     a4, #0
sm_loop:
        CMP     a2, a3, ASL #0
        BLS     sm_shifted0mod8
        CMP     a2, a3, ASL #1
        BLS     sm_shifted1mod8
        CMP     a2, a3, ASL #2
        BLS     sm_shifted2mod8
        CMP     a2, a3, ASL #3
        BLS     sm_shifted3mod8
        CMP     a2, a3, ASL #4
        BLS     sm_shifted4mod8
        CMP     a2, a3, ASL #5
        BLS     sm_shifted5mod8
        CMP     a2, a3, ASL #6
        BLS     sm_shifted6mod8
        CMP     a2, a3, ASL #7
        MOVHI   a3, a3, ASL #8
        BHI     sm_loop
sm_loop2:
        CMP     a2, a3, ASL #7
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #7
        CMP     a2, a3, ASL #6
sm_shifted6mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #6
        CMP     a2, a3, ASL #5
sm_shifted5mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #5
        CMP     a2, a3, ASL #4
sm_shifted4mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #4
        CMP     a2, a3, ASL #3
sm_shifted3mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #3
        CMP     a2, a3, ASL #2
sm_shifted2mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #2
        CMP     a2, a3, ASL #1
sm_shifted1mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #1
        CMP     a2, a3, ASL #0
sm_shifted0mod8:
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #0
        CMP     a1, a3, LSR #1
        MOVLS   a3, a3, LSR #8
        BLS     sm_loop2
        MOV     a1, a4
        TST     ip, #0x40000000
        RSBNE   a1, a1, #0
        TST     ip, #0x80000000
        RSBNE   a2, a2, #0
	MOV	a1, a2		// Devolver resto
        MOV	pc, lr

	.globl	__udivsi3
__udivsi3:
//
// Unsigned divide of a2 by a1: returns quotient in a1, remainder in a2.
// Also destroys a3, a4 and ip
	MOV	ip, a2
	MOV	a2, a1
	MOV	a1, ip

        MOVS    a3, a1
        BEQ     dividebyzero
        MOV     a4, #0
        MOV     ip, #0x80000000
        CMP     a2, ip
        MOVLO   ip, a2
u_loop:
        CMP     ip, a3, ASL #0
        BLS     u_shifted0mod8
        CMP     ip, a3, ASL #1
        BLS     u_shifted1mod8
        CMP     ip, a3, ASL #2
        BLS     u_shifted2mod8
        CMP     ip, a3, ASL #3
        BLS     u_shifted3mod8
        CMP     ip, a3, ASL #4
        BLS     u_shifted4mod8
        CMP     ip, a3, ASL #5
        BLS     u_shifted5mod8
        CMP     ip, a3, ASL #6
        BLS     u_shifted6mod8
        CMP     ip, a3, ASL #7
        MOVHI   a3, a3, ASL #8
        BHI     u_loop
u_loop2:
u_shifted7mod8:
        CMP     a2, a3, ASL #7
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #7
u_shifted6mod8:
        CMP     a2, a3, ASL #6
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #6
u_shifted5mod8:
        CMP     a2, a3, ASL #5
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #5
u_shifted4mod8:
        CMP     a2, a3, ASL #4
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #4
u_shifted3mod8:
        CMP     a2, a3, ASL #3
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #3
u_shifted2mod8:
        CMP     a2, a3, ASL #2
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #2
u_shifted1mod8:
        CMP     a2, a3, ASL #1
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #1
u_shifted0mod8:
        CMP     a2, a3, ASL #0
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #0
        CMP     a1, a3, LSR #1
        MOVLS   a3, a3, LSR #8
        BLS     u_loop2
        MOV     a1, a4
        MOV	pc, lr

	.globl	__umodsi3
__umodsi3:
//
// Unsigned divide of a2 by a1: returns quotient in a1, remainder in a2.
// Also destroys a3, a4 and ip
	MOV	ip, a2
	MOV	a2, a1
	MOV	a1, ip

        MOVS    a3, a1
        BEQ     dividebyzero
        MOV     a4, #0
        MOV     ip, #0x80000000
        CMP     a2, ip
        MOVLO   ip, a2
um_loop:
        CMP     ip, a3, ASL #0
        BLS     um_shifted0mod8
        CMP     ip, a3, ASL #1
        BLS     um_shifted1mod8
        CMP     ip, a3, ASL #2
        BLS     um_shifted2mod8
        CMP     ip, a3, ASL #3
        BLS     um_shifted3mod8
        CMP     ip, a3, ASL #4
        BLS     um_shifted4mod8
        CMP     ip, a3, ASL #5
        BLS     um_shifted5mod8
        CMP     ip, a3, ASL #6
        BLS     um_shifted6mod8
        CMP     ip, a3, ASL #7
        MOVHI   a3, a3, ASL #8
        BHI     um_loop
um_loop2:
um_shifted7mod8:
        CMP     a2, a3, ASL #7
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #7
um_shifted6mod8:
        CMP     a2, a3, ASL #6
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #6
um_shifted5mod8:
        CMP     a2, a3, ASL #5
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #5
um_shifted4mod8:
        CMP     a2, a3, ASL #4
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #4
um_shifted3mod8:
        CMP     a2, a3, ASL #3
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #3
um_shifted2mod8:
        CMP     a2, a3, ASL #2
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #2
um_shifted1mod8:
        CMP     a2, a3, ASL #1
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #1
um_shifted0mod8:
        CMP     a2, a3, ASL #0
        ADC     a4, a4, a4
        SUBHS   a2, a2, a3, ASL #0
        CMP     a1, a3, LSR #1
        MOVLS   a3, a3, LSR #8
        BLS     um_loop2
        MOV     a1, a4

	MOV	a1, a2		// Return remainder
        MOV	pc, lr


#ifdef DIV10_OPTIMUS
	.globl	__rt_sdiv10
__rt_sdiv10:
//
// Fast signed divide by 10: dividend in a1, divisor in a2.
// Returns quotient in a1, remainder in a2.
// Also destroys a3 and a4.
// Quotient is truncated (rounded towards zero).

        MOVS    a4, a1
        RSBMI   a1, a1, #0
        MOV     a2, a1
        MOV     a1, a1, LSR #1
        ADD     a1, a1, a1, LSR #1
        ADD     a1, a1, a1, LSR #4
        ADD     a1, a1, a1, LSR #8
        ADD     a1, a1, a1, LSR #16
        MOV     a1, a1, LSR #3
        ADD     a3, a1, a1, ASL #2
        SUB     a2, a2, a3, ASL #1
        CMP     a2, #10
        ADDGE   a1, a1, #1
        SUBGE   a2, a2, #10
        CMP     a4, #0
        RSBMI   a1, a1, #0
        RSBMI   a2, a2, #0
        MOV	pc, lr

	.globl	__rtudiv10
__rt_udiv10:
//
// Fast unsigned divide by 10: dividend in a1, divisor in a2.
// Returns quotient in a1, remainder in a2.
// Also destroys a3.
//
// Calculate x / 10 as (x * 2**32/10) / 2**32.
// That is, we calculate the most significant word of the double-length
// product. In fact, we calculate an approximation which may be 1 off
// because we've ignored a carry from the least significant word we didn't
// calculate. We correct for this by insisting that the remainder < 10
// and by incrementing the quotient if it isn't.

        MOV     a2, a1
        MOV     a1, a1, LSR #1
        ADD     a1, a1, a1, LSR #1
        ADD     a1, a1, a1, LSR #4
        ADD     a1, a1, a1, LSR #8
        ADD     a1, a1, a1, LSR #16
        MOV     a1, a1, LSR #3
        ADD     a3, a1, a1, ASL #2
        SUB     a2, a2, a3, ASL #1
        CMP     a2, #10
        ADDGE   a1, a1, #1
        SUBGE   a2, a2, #10
        MOV	pc, lr
#endif

dividebyzero:
	MOV	r0, #0
	MOV	r1, #0
	MOV	pc, lr

	.globl	reset
reset:
	MOV	pc, #0x41000000

