
/***************************************************************************
    libBIOS - A portable library for developing an OpenSource BIOS
    Copyright (C) 2001 - David Aparicio

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author can be reached at 'daparic@terra.es'
****************************************************************************/
// 
#include "emOS.h"
#include "macros.h"

//-------------------------------------------------------------------------
//---- MultiUser routines. This mode is MULTITASK (SPV kernel, USR tasks)
//----     1) SP initialization for FIQ, IRQ and USR
//----     2) INTERRUPTS initialization

//--------------------------------- Estructura de contexto de registros
//    bit 31 - NZCVxxxx.xxxxxxxx.xxxxxxxx.IFTmmmmm - bit 0
//    I: Enable(0) / Disable(1) IRQ
//    F: Enable(0) / Disable(1) FIQ
//mmmmm: 10000 user
//       10001 FIQ   (save r8 -r14, spsr=PSR, r14=PC)
//       10010 IRQ   (save r13-r14, spsr=PSR, r14=PC)
//       10011 super (save r13-r14, spsr=PSR, r14=PC)
//       10111 abort (save r13-r14, spsr=PSR, r14=PC)
//       11011 undef (save r13-r14, spsr=PSR, r14=PC)
//--- En [0x00] RES I:1, F:1, m:10011
//             no se retorna
//--- En [0x08] SWI I:1, F:x, m:10011, R14=interruptedPC, SPSR=interruptedPSR
//             se retorna con [MOVS PC,R14]
//--- En [0x18] IRQ I:1, F:x, m:10010, R14=interruptedPC, SPSR=interruptedPSR
//             se retorna con [SUBS PC,R14,#4]
//--- En [0x1C] FIQ I:1, F:1, m:10001, R14=interruptedPC, SPSR=interruptedPSR
//             se retorna con [SUBS PC,R14,#4]

beginMulti:	.globl beginMulti
	MOV	r2, #ARMCSR_BASE
	LDR	r7, [r2, #OFS_MBOX1]	// ramsize

	MOV	r8, lr			// Save RET in a non-banked register
					// we enter as SVR, but return as USR
	// Establecer pila de FIQ32 mode
	MOV	r0, #csrIRQf+csrFIQf+csrFIQm
	MSR	cpsr_c, r0
	SUB	sp, r7, #0x00000100	// Fin de RAM menos 256 B

	// Establecer pila de IRQ32 mode
	MOV	r0, #csrIRQf+csrIRQm
	MSR	cpsr_c, r0
	SUB	sp, r7, #0x00000800	// Fin de RAM menos 2 KB

	// Pasar a modo USR
	MOV	r0, #csrUSRm
	MSR	cpsr_c, r0
	SUB	sp, r7, #0x00000400	// Pila inicial, solo para KNLreset

	// Habilitar interrupciones linea serie (DEBUG)
	MOV	r0, #FLG_IRQ_RX
	STR	r0, [r2, #OFS_FIQ_ENABLE]

	// inicializacion de alto nivel	(elegir primer RUNNING)
	// No debe retornar, sino que "pasa el contexto" a TASK[0]
	LDR	r0, [r2, #OFS_MBOX0]	// env_t (arg0)
	MOV	pc, r8

haltMulti: .globl	haltMulti
	// Establecer SVC mode
	MOV	r0, #csrIRQf+csrSVCm
	MSR	cpsr_c, r0
	
	// Desabilitar todas las interrupciones
	MOV	r0, #0xFFFFFFFF
	STR	r0, [r1, #OFS_IRQ_DISABLE]
	STR	r0, [r1, #OFS_FIQ_DISABLE]

	// Saltar a la zona de descarga (como si hubiera sido despues de POST)
	MOV	r0, #ADDR_LOAD		// JUMP to DOWNLOAD ZONE
	MOV	pc, lr

//--------------------------------- TRAPS -----------------------------------
SWItrap:	.globl SWItrap	// *** IRQ disabled ***
	// --- Construccion argumentos y llamada C
	LDR	r3, [lr, #-4]
	BIC	r3, r3, #0xFF000000	// SWI number (arg4)

	// --- Llamada a SYSCALL
	BL	KNLswi

	// --- Comprobar si hay cambio de contexto tras la SYSCALL
	LDR	r2, =sys
	LDR	r1, [r2, #SYS_CALLING]
	LDR	r2, [r2, #SYS_RUNNING]
	CMP	r1, r2
	BEQ	noctx

	// --- Retorno con cambio de contexto
	POINT	lr, SYS_CALLING
	saveTCBregs
	LDR	r0, [lr, #O_R14]	// El PC y LR de usuario coinciden
	STR	r0, [lr, #O_PC]
	POINT	lr, SYS_RUNNING
	loadTCBregs
	LDR	lr, [lr, #O_PC]
	MOVS	pc, lr

noctx:	// --- Retorno sin cambio de contexto
	POINT	lr, SYS_RUNNING
	LDR	r0, [lr, #O_R0 ]	// Recuperar valor de retorno
	pushULR
	pop	lr
	MOVS	pc, lr

//---------
IRQtrap:	.globl IRQtrap
	// --- Apilar PC de usuario
	SUB	lr, lr, #4
	push	lr
	// --- Salvar contexto actual
	POINT	lr, SYS_RUNNING
	saveTCBregs
	// --- Salvar PC interrumpido
	pop	r0
	STR	r0, [lr, #O_PC]

	// --- Preparar pila y llamar a C
	BL	KNLirq

	// --- Cargar TIC para siguiente interrupcion
	MOV	r3, #ARMCSR_BASE
	STR	r3, [r3, #OFS_TIMER1_CLR]

	// --- Cargar contexto actual
	POINT	lr, SYS_RUNNING
	loadTCBregs
	LDR	lr, [lr, #O_PC]
	MOVS	pc, lr

//---------
FIQtrap:	.globl FIQtrap
	SUB	lr, lr, #4
	push	lr
	push	r0
	push	r1
	push	r2
	push	r3
	push	r4
	push	r5
	push	r6
	push	r7
	MOV	r0, lr
	bl	stub_debug
	//	Flush serial characters
	B	1f
0:	bl	getc
1:	bl	keysleft
	cmp	r0, #0
	bne	0b
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	pop	r3
	pop	r2
	pop	r1
	pop	r0
	pop	lr
	MOVS	pc, lr

//----------------
	.globl	initRegs
initRegs:		// (TCB_PTR, ARG0, PC, SP)
		// R0: (u32 *)   Contexto a inicializar
		// R1: (pid_t)   pid inicial (arg0)
		// R2: (void *)  Program counter inicial (principio de tarea)
		// R3: (u32 *)   Stack pointer inicial (final de espacio stack)
	//-- save ARG0
	STR	r1, [r0, #O_R0]
	//-- save PC
	STR	r2, [r0, #O_R14]
	STR	r2, [r0, #O_PC]
	//-- save SP
	MOV	r2, #STACK_SIZE	// r3 = r3 + (STACKSIZE-1)<<2
	SUB	r2, r2, #1
	MOV	r2, r2, lsl #2
	ADD	r3, r3, r2
	STR	r3, [r0, #O_R13]
	//-- init another regs
	MOV	r3, #csrUSRm	// modo USR, IRQ=on, FIQ=on
	STR	r3, [r0, #O_FL]
	MOV	r3, #0
	STR	r3, [r0, #O_R1]
	STR	r3, [r0, #O_R2]
	STR	r3, [r0, #O_R3]
	STR	r3, [r0, #O_R4]
	STR	r3, [r0, #O_R5]
	STR	r3, [r0, #O_R6]
	STR	r3, [r0, #O_R7]
	STR	r3, [r0, #O_R8]
	STR	r3, [r0, #O_R9]
	STR	r3, [r0, #O_R10]
	STR	r3, [r0, #O_R11]
	STR	r3, [r0, #O_R12]
	MOV	pc, lr

	.globl	CPUdump
	// Imprime ARG0, ARG1 y ARG2
CPUdump:
	STMDB	sp!, { r4-r12, lr }
	MOV	r10, r0
	MOV	r11, r1
	MOV	r12, r2
	ldr	r0, =sys
	ldr	r0, [r0, #SYS_TOTALTIC]
	printHEXbl	r0
	printHEXbl	r10
	printHEXbl	r11
	printHEXnl	r12
	LDMIA	sp!, { r4-r12, pc }

