
/***************************************************************************
    libBIOS - A portable library for developing an OpenSource BIOS
    Copyright (C) 2001 - David Aparicio

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author can be reached at 'daparic@terra.es'
****************************************************************************/
// 
#include "libBIOS.h"
#include "ali1543.h"
#include "dc21285.h"

//-- Macros for peek and poke on CSR space in a specific (device dependant) way.
//-- I hope you have a decent datasheet if you intend to use this...
#define CSR_OFFS8(CSR, OFFS)  ( u8 *)((char *)CSR + OFFS)
#define CSR_OFFS16(CSR, OFFS) (u16 *)((char *)CSR + OFFS)
#define CSR_OFFS32(CSR, OFFS) (u32 *)((char *)CSR + OFFS)

#define PCI_IO_ENABLE		0x0001
#define PCI_MEM_ENABLE		0x0002
#define PCI_MASTER_ENABLE	0x0004
#define PCI_MEM_WRINVALID	0x0010
#define PCI_PERR_ENABLE 	0x0020
#define PCI_SERR_ENABLE 	0x0080
#define PCI_FAST_B2B		0x0200

//--------- IRQ mapping for CATS -(ISA compatible PIC) ------------------------
//--------- these codes stands for Linux 2.2 compatibility, due to Cyclone/CATS
//--------- "original" assignments. What we do is mimetize original mapping so
//--------- no new "architecture" is needed. We hope this to change some day 
//--- If anybody can assure that this is not any longer necessary, I will be
//--- VERY pleased to remove it.

//---- We identify device by IDSEL and assign it a FIXED IRQ for routing.
//---- We also have INTA to INTD lines directly routed to FootBridge, but this
//---- is not used by now (as of 2.2 kernel)
const s16 irq_slot[] = {
	/*  0   1   2   3   4   5   6   7  SYSTEM */
	   -1, -1, -1, -1, -1, -1, -1,  0,
	/*  8   9  10  11  12  13  14  15  PCI ADD ON SLOTS (8, 9, 10 y 11) */
	    1,  2,  3,  4, -1, -1, -1, -1,
	/* 16  17  18  19  20  21          SPECIAL: IDE 16, PMU 17, USB 20 */
	  255,  0, -1, -1,139, -1
	};

int pci_hostInit(pciFrame_t *pfm)
{
	env_t *env = (void *)(*CSR_MBOX0);
	u32 rammask, rommask, hostmode, cntl;

	cntl = *CSR_SA110_CNTL;
	//--- These values should be here from 1st step procedure...

	switch(env->ramsize)
	{
		case SZ_256K: rammask = 0x00000000; break;
		case SZ_512K: rammask = 0x00040000; break;
		case SZ_1M  : rammask = 0x000C0000; break;
		case SZ_2M  : rammask = 0x001C0000; break;
		case SZ_4M  : rammask = 0x003C0000; break;
		case SZ_8M  : rammask = 0x007C0000; break;
		case SZ_16M : rammask = 0x00FC0000; break;
		case SZ_32M : rammask = 0x01FC0000; break;
		case SZ_64M : rammask = 0x03FC0000; break;
		case SZ_128M: rammask = 0x07FC0000; break;
		case SZ_256M: rammask = 0x0FFC0000; break;
		default:      rammask = 0x8FFC0000; env->ramsize = 0; break;
	}
	switch(env->romsize)
	{
		case SZ_2M  : rommask = 0x00100000; break;
		case SZ_4M  : rommask = 0x00300000; break;
		case SZ_8M  : rommask = 0x00700000; break;
		default:      rommask = (env->romsize<SZ_2M) ? 0 : 0x00F00000;
			      env->romsize =(env->romsize<SZ_2M) ?SZ_1M :SZ_16M;
                              break;
	}
	hostmode = (cntl&0x80000000)>0;
	*CSR_PCI_INTMASK    = 0x0C;
	*CSR_DOORBELL_PCI   = 0;
	*CSR_DOORBELL_SA110 = 0;

	// PCI configure mask & offset
	*CSR_SDRAMBASEOFFSET= 0;
	*CSR_SDRAMBASEMASK  = rammask;
	*CSR_ROMBASEMASK    = rommask;
	*CSR_CSRBASEOFFSET  = 0;
	*CSR_CSRBASEMASK    = 0;
	*CSR_PCIADDR_EXTN   = 0;	// PCI direcciona en 0x00000000
	*CSR_PCI_INTLINE    = 1;	// INTA
	cntl |= hostmode
			? 0x00000200		// de-assert pci_rst_l
			: 0x60000000;		// Enable XBUS chsel 1 & 2
	if(hostmode && pfm)
	{
		*CSR_PCICMD	    = 0;
		*CSR_DOORBELL_SETUP = 0;
		// Assign resources for this host
		*CSR_PCICSRIOBASE	= pfm->min[PCI_GRPIO ];
		pfm->min[PCI_GRPIO ] += SZ_4K;	// Roundup for I/O in bridge
		*CSR_PCICSRBASE		= pfm->min[PCI_GRPMEM];
		pfm->min[PCI_GRPMEM] += SZ_1M;	// Roundup for MEM in bridge
		*CSR_PCIROMBASE		= pfm->min[PCI_GRPMEM];
		pfm->min[PCI_GRPMEM] += env->romsize;
		*CSR_PCISDRAMBASE	= pfm->min[PCI_GRPRAM];
		pfm->min[PCI_GRPRAM] += env->ramsize;

		*CSR_PCICMD	    =	PCI_IO_ENABLE |
					PCI_MEM_ENABLE | 
					PCI_MASTER_ENABLE |
					PCI_MEM_WRINVALID;
	}
	*CSR_SA110_CNTL     = cntl|1;	// PCI inicialization complete
	if(hostmode)	// Wait 300 ms if we deaserted reset line (PCI)
		wait(300, 0);
	else		// Not PCI host. We suppose it will be a EBSA285
	{
		volatile u8 *jumper;
		u8 xleds;

		// Enable XBUS (leds)
		*CSR_XBUS_CYCLE    = 0x100016DB;
		*CSR_XBUS_IOSTROBE = 0xFCFCFCFC;
		jumper = (u8 *)0x40012000; xleds = 7;

		*CSR_MBOX2 = 0xFFFFFFFF;
		while(*CSR_MBOX2 == 0xFFFFFFFF)
		{
			*jumper = xleds;
			wait(250, 0);
			xleds = xleds ? 0 : 7;
		}
		//--- MBOX2 has entry address to jump.
		//--- We suppose it will be ADDR_LOAD
	}
	return hostmode;
}

//--- Minimum hardware inicialization (device scanning - Detect requests)
//--- For host (leading EBSA285), return 1; [proceed with POST procedure]
//--- Any other case (not EBSA or not motherboard), return 0; [don't proceed]
int pci_devScan(csr_t *csr)
{
	u8 bus, dev, fun;

	pci_cnvCsrPar(csr, &bus, &dev, &fun);
	if(bus+dev+fun==0)
		return pci_hostInit(NULL);
	else switch(PCI_DEVICE(csr))
	{
	case 0x10B91533:	// PCI/ISA (ALi) bridge
		// Enable  IDE interface (in PCI/ISA bridge)
		*CSR_OFFS8(csr, 0x58) = 0x4C;
		// Disable USB interface (in PCI/ISA bridge)
//		*CSR_OFFS8(csr, 0x53) = 0x40;
		// Disable PMU interface (in PCI/ISA bridge)
		*CSR_OFFS8(csr, 0x5F) = 0x0E;
		// Wait 30 ms
		wait(30, 0);
		break;
	default:
		break;
	}
	return 1;
}

//--- Full hardware inicialization (device assignment - Accomplish requests)
void pci_devInit(csr_t *csr, u32 addr[], u32 low[], u32 hig[])
{
	u8 bus, dev, fun;
	int i;

	pci_cnvCsrPar(csr, &bus, &dev, &fun);
	csr->command = 0;
	if(bus+dev+fun == 0)
		*CSR_DOORBELL_SETUP = 0;
	else switch(PCI_DEVICE(csr))
	{
	case 0x10B91533:	// PCI/ISA (ALi) bridge
#define M1543_PNPW(A, D) { \
	*PCIPORT(M1543_PNPIDX) = (A);	\
	*PCIPORT(M1543_PNPDAT) = (D);	\
	}
		// ISA PnP: Enter CONFIG mode
		*PCIPORT(M1543_PNPCFG) = 0x51;
		*PCIPORT(M1543_PNPCFG) = 0x23;

		M1543_SETDEV(M1543_DEVKBD);	// PS/2, KBD: Disable
		M1543_PNPW(0x30, 0x00);
	
		M1543_SETDEV(M1543_DEVFDC);	// FLOPPY: Disable
		M1543_PNPW(0x30, 0x00);

		M1543_SETDEV(M1543_DEVUA0);	// UART0: Disable
		M1543_PNPW(0x30, 0x00);
	
		M1543_SETDEV(M1543_DEVUA1);	// UART1: Disable
		M1543_PNPW(0x30, 0x00);
	
		M1543_SETDEV(M1543_DEVPP0);	// PP0: Enable
		M1543_PNPW(0x30, 0x01);
		M1543_PNPW(0x60, 0x03);		// Decode HI-addr 0x0300
		M1543_PNPW(0x61, 0x00);		// Decode LO-addr 0x0300
		M1543_PNPW(0x70, 0x00);		// No IRQ
		M1543_PNPW(0x74, 0x04);		// No DMA
		M1543_PNPW(0xF0, 0x0D);		// EPP1.7
		M1543_PNPW(0xF1, 0x00);		// 24MHz, no burst
	
		//-- ISA PnP: Enter RUN mode
		*PCIPORT(M1543_PNPCFG) = 0xBB;
		wait(10, 0);
//		*PCIPORT(PP0_ECPCFG) = 0x80;  // PP0: Enable EPP

		*CSR_OFFS8(csr, 0x40) = 0x30; // PCI control
		*CSR_OFFS8(csr, 0x41) = 0x13; // ISA I/O recovery
		*CSR_OFFS8(csr, 0x42) = 0x08;
		*CSR_OFFS8(csr, 0x43) = 0xEA;

		*CSR_OFFS8(csr, 0x58) = 0x4C; // ISA/IDE interface control
		*CSR_OFFS8(csr, 0x44) = 0x50; // IDE0: no ISA/IRQ
		*CSR_OFFS8(csr, 0x75) = 0x50; // IDE1: no ISA/IRQ
		break;
	case 0x10B95229:	// ALi IDE
		*CSR_OFFS8(csr, 0x51) = 0xF0; // reset IDE
#ifndef IDE_COMPATIBLE
		*CSR_OFFS8(csr, 0x09) = 0xF5; // Native mode only
#endif
		*CSR_OFFS8(csr, 0x4D) = 0x80; // Index 0x09 readonly
		*CSR_OFFS8(csr, 0x4F) = 0x38; // CONFreg
		*CSR_OFFS8(csr, 0x50) = 0x0B; // CONFreg
		*CSR_OFFS8(csr, 0x53) = 0x03; // CDROM PIO & DMA support
		*CSR_OFFS8(csr, 0x54) = 0x00; // FIFO chn1
		*CSR_OFFS8(csr, 0x55) = 0x00; // FIFO chn2
		*CSR_OFFS8(csr, 0x56) = 0x88; // IDE0 UltraDMA/33 enabled
		*CSR_OFFS8(csr, 0x57) = 0x88; // IDE1 UltraDMA/33 enabled
		break;
	case 0x10B97101:	// ALi PMU
		*CSR_OFFS8(csr, 0x48) = 0x60;
		*CSR_OFFS8(csr, 0x49) = 0x08;
		*CSR_OFFS8(csr, 0x7C) = 0x10;
		*CSR_OFFS8(csr, 0xB3) = 0x70;
		*CSR_OFFS8(csr, 0xB5) = 0x06;
		*CSR_OFFS8(csr, 0xE0) = 0x01;
		*CSR_OFFS8(csr, 0xE2) = 0x20;
		break;
	case 0x10B95237:	// ALi USB
		*CSR_OFFS8(csr, 0x42) = 0x1F;	// USB?? delante o detras
		break;
	default:	// generic device, nothing to do
		break;
	}

	csr->latency = 32;	// Might some guru choose a proper value?

	//--- MEM & I/O assignment
	if((csr->header & 0x7F) == 0)
	{
		for(i=0; i<6; i++)
			if(addr[i] != 0xFFFFFFFF)
				csr->u.hdr0.baseaddr[i] = addr[i];
		if(addr[6] != 0xFFFFFFFF)
			csr->u.hdr0.rom_base = addr[i] | 1;
	}
	else
	{
		for(i=0; i<2; i++)
			if(addr[i] != 0xFFFFFFFF)
				csr->u.hdr1.baseaddr[i] = addr[i];
		if(addr[6] != 0xFFFFFFFF)
			csr->u.hdr1.rom_base = addr[i] | 1;

		// Assign FORWARDING maps for bridge devices...
		// Set I/O model to 16 bits, MEM model to 32 bits
		csr->u.hdr1.hi32_rambase = 0;
		csr->u.hdr1.hi32_ramsize = 0;
		csr->u.hdr1.hi16_iobase  = 0;
		csr->u.hdr1.hi16_iosize  = 0;
		// I/O:  4 significant bits ( 8 bit in reg, 16 bit addr)
		if(low[PCI_GRPIO] < hig[PCI_GRPIO])
		{
			csr->u.hdr1.iobase  = low[PCI_GRPIO ]>> 8;
			csr->u.hdr1.iosize  = hig[PCI_GRPIO ]>> 8;
		}
		else
		{
			csr->u.hdr1.iobase  = 0x00;
			csr->u.hdr1.iosize  = 0x00;
		}
		// MEM: 12 significant bits (16 bit in reg, 32 bit addr)
		if(low[PCI_GRPMEM] < hig[PCI_GRPMEM])
		{
			csr->u.hdr1.membase = low[PCI_GRPMEM]>>16;
			csr->u.hdr1.memsize = hig[PCI_GRPMEM]>>16;
		}
		else
		{
			csr->u.hdr1.membase = 0x0000;
			csr->u.hdr1.memsize = 0x0000;
		}
		// RAM: 12 significant bits (16 bit in reg, 32 bit addr)
		if(low[PCI_GRPRAM] < hig[PCI_GRPRAM])
		{
			csr->u.hdr1.rambase = low[PCI_GRPRAM]>>16;
			csr->u.hdr1.ramsize = hig[PCI_GRPRAM]>>16;
		}
		else
		{
			csr->u.hdr1.rambase = 0x0000;
			csr->u.hdr1.ramsize = 0x0000;
		}
	}
#ifdef DEPURACION
	printk("[%i,%2i] %08X %08X %08X %08X %08X %08X %08X\n",
			bus, dev, addr[0], addr[1], addr[2],
			addr[3], addr[4], addr[5], addr[6]);
	if((csr->header & 0x7F) != 0)
		printk( "            %i>%i %08X-%08X %08X-%08X %08X-%08X\n",
					csr->u.hdr1.busE, csr->u.hdr1.busI,
					low[PCI_GRPIO ], hig[PCI_GRPIO ],
					low[PCI_GRPMEM], hig[PCI_GRPMEM],
					low[PCI_GRPRAM], hig[PCI_GRPRAM]
					);
#endif
	//--- IRQ assignment
	if((csr->header & 0x7F) == 0)
		if(csr->u.hdr0.ipin)
			csr->u.hdr0.irq = irq_slot[pci_getNumSlot(csr)];
		else
			csr->u.hdr0.irq = 0xFF;
	else
		if(csr->u.hdr1.ipin)
			csr->u.hdr1.irq = irq_slot[pci_getNumSlot(csr)];
		else
			csr->u.hdr1.irq = 0xFF;

	// The easy way: We enable all capabilities. Devices not capable
	//   of some of this, will have a "not-writable" bit on proper
	//   property.
	csr->command =	PCI_MASTER_ENABLE |
			PCI_IO_ENABLE |
			PCI_MEM_ENABLE |
			PCI_MEM_WRINVALID;
}
