#include "mon.h"
#include "dc21285.h"

typedef struct {
			// definition of this device
	char	*name;
			// size of device (in bytes)
	u32	total_size;
			// size of protected boot-sector (in bytes)
			// 0 if this device doesn't have this feature
	u32	boot_size;
			// size of block (in bytes)
			// 1 if this device doesn't have sectors
	u32	block_size;
	// devuelve kbytes de capacidad o 0 si no se detecta
	int	(*detect)(u32 romaddr);
	// realiza la proteccion del bootsector
	int	(*lock)(u32 romaddr);
	// devuelve bytes escritos, o 0 si error
	// si data==NULL, devuelve tamao de sector, y no hace nada
	int	(*write)(u32 romaddr, volatile u8 *data, u32 size);
	} flash_t;

static const flash_t  dev[];
static const flash_t *device = NULL;

INLINE u8 R8(u32 addr)
	{	return *(volatile u8 *)addr; }

INLINE void W8(u32 addr, u8 val)
	{ *CSR_ROMWRITEREG=(addr&3); *(volatile u8 *)(addr&0xFFFFFFFC)=val; }

INLINE u32 R32(u32 base, u32 offs)
	{	return *(volatile u32 *)(base+offs); }

INLINE void W32(u32 base, u32 offs, u32 val)
	{	R32(base, 0); *(volatile u32 *)(base+offs) = val; }

void wait_togglebit(u32 addr)
{
	u8  tmp0, tmp1;

	printk("togglebit %08X\n", addr);
	// Toggle bit algorithm: Espera escritura realizada
	tmp0 = R8(addr);
	while(1)
	{
		tmp1 = R8(addr);

		if((tmp0 ^ tmp1) & 0x40)
			tmp0 = tmp1;
		else
			break;
	}
}

INLINE int wait_truedata(u32 addr, u8 data)
{
	int counter = 100;

	wait_ns(30);
	while(R8(addr) != data)
	{
		wait_ns(20);
		if(!counter--) return 0;
	}
	return 1;
}

#define DEV_ERROR  0
#define DEV_LOCKED 1
#define DEV_NOLOCK 2

//--SST20---------------------------------------------------------------------
static int sst20detect(u32 base)
{
	u8 id1, id2;

	// ID entry
	W8(base+0x5555, 0xAA); W8(base+0x2AAA, 0x55); W8(base+0x5555, 0x90);
	wait(2, 0);

	id1 = R8(base); id2 = R8(base);

	// ID exit
	W8(base+0x5555, 0xAA); W8(base+0x2AAA, 0x55); W8(base+0x5555, 0xF0);
	wait(2, 0);
	return id1==0xBF && id2==0x12 ? DEV_NOLOCK : DEV_ERROR;
}

static int sst20write(u32 base, volatile u8 *data, u32 len)
{
	if(!device || len > device->total_size) return 0;

	printk("Programming %s\n", device->name);
	while(len)
	{
		u32 offs;
		u16 blen = len > device->block_size ? device->block_size : len;
		// Graba un bloque de 128 caracteres (de RAM a flash)
		W8(base+0x5555, 0xAA);
		W8(base+0x2AAA, 0x55);
		W8(base+0x5555, 0xA0);

		for(offs = 0; blen; offs++, data++, len--, blen--)
			W8(base+offs, *data);
		wait(1, 0);	// Esperar Tblco (ciclo de volcado de bloque)
//		wait_truedata(base+offs-1, *(data-1));
		wait_togglebit(base);
	}
	// Disable write mode...
	W8(base+0x5555, 0xAA);
	W8(base+0x2AAA, 0x55);
	W8(base+0x5555, 0xA0);
	wait(10, 0);
	return 1;
}

//--AT020---------------------------------------------------------------------
static int  at20detect(u32 base)
{
	u8 id1, id2, id3;

	// ID entry
	W8(base+0x5555, 0xAA); W8(base+0x2AAA, 0x55); W8(base+0x5555, 0x90);
	wait(20, 0);

	id1 = R8(base+0); id2 = R8(base+1); id3 = R8(base+2);

	// ID exit
	W8(base+0x5555, 0xAA); W8(base+0x2AAA, 0x55); W8(base+0x5555, 0xF0);
	wait(20, 0);
	if(id1==0x1F && id2==0xBA)
		return id3==0xFE ? DEV_NOLOCK : DEV_LOCKED;
	else
		return DEV_ERROR;
}

static int  at20lock(u32 base)
{
	if(!device) return 0;
	// OTP bootsector cycle
	W8(base+0x5555, 0xAA);
	W8(base+0x2AAA, 0x55);
	W8(base+0x5555, 0x80);
	W8(base+0x5555, 0xAA);
	W8(base+0x2AAA, 0x55);
	W8(base+0x5555, 0x40);
	// Select LOW ADDRESS boot (as per EBSA285 architecture)
	W8(base+0x0000, 0x00);
	wait(20, 0);
	return 1;
}

static int  at20write(u32 base, volatile u8 *data, u32 len)
{
	if(!device || len > device->total_size) return 0;

	printk("Programming %s\n", device->name);
	while(len)
	{
		u32 offs;
		u16 blen = len > device->block_size ? device->block_size : len;
		// Graba un bloque de 256 caracteres (de RAM a flash)
		W8(base+0x5555, 0xAA);
		W8(base+0x2AAA, 0x55);
		W8(base+0x5555, 0xA0);

		for(offs=0; blen; offs++, data++, len--, blen--)
			W8(base+offs, *data);
		wait(1, 0);	// Esperar Tblco (ciclo de volcado de bloque)
//		wait_truedata(base+offs-1, *(data-1));
		wait_togglebit(base);
	}
	return 1;
}

//--AT040---------------------------------------------------------------------
static int  at40detect(u32 base)	// 0 si ok, -1 si no encuentra el chip
{
	u8 id1, id2, id3;

	// ID entry
	W8(base+0x5555, 0xAA); W8(base+0x2AAA, 0x55); W8(base+0x5555, 0x90);
	wait(20, 0);

	id1 = R8(base+0); id2 = R8(base+1); id3 = R8(base+2);

	// ID exit
	W8(base+0x5555, 0xAA); W8(base+0x2AAA, 0x55); W8(base+0x5555, 0xF0);
	wait(20, 0);
	if(id1==0x1F && id2==0x13)
		return (id3&1)==0 ? DEV_NOLOCK : DEV_LOCKED;
	else
		return DEV_ERROR;
}

static int  at40lock(u32 base)
{
	if(!device) return 0;
	// OTP bootsector cycle
	W8(base+0x5555, 0xAA);
	W8(base+0x2AAA, 0x55);
	W8(base+0x5555, 0x80);
	W8(base+0x5555, 0xAA);
	W8(base+0x2AAA, 0x55);
	W8(base+0x5555, 0x40);
	wait(1000, 0);
	return 1;
}

static int  at40write(u32 base, volatile u8 *data, u32 len)
{
	u32 offs;

	if(len > device->total_size) return 0;

	printk("erasing %i bytes (about 10 seg)...", len);

	// Erase cycle
	W8((base&0xFFFF0000)|0x5555, 0xAA);
	W8((base&0xFFFF0000)|0x2AAA, 0x55);
	W8((base&0xFFFF0000)|0x5555, 0x80);
	W8((base&0xFFFF0000)|0x5555, 0xAA);
	W8((base&0xFFFF0000)|0x2AAA, 0x55);
	W8((base&0xFFFF0000)|0x5555, 0x10);
	wait(12000, 0);
//	wait_togglebit(base);	// hasta 10s
	printk("writing...");
	for(offs = 0; offs<len; offs++, data++)
	{
		if((offs%128)==0) printk("%6i\b\b\b\b\b\b", offs);
		while(1)
		{
			// Byte-program cycle
			W8((base&0xFFFF0000)|0x5555, 0xAA);
			W8((base&0xFFFF0000)|0x2AAA, 0x55);
			W8((base&0xFFFF0000)|0x5555, 0xA0);
			W8(base+offs, *data);
			if(wait_truedata(base+offs, *data)) break;
		}
//		wait_togglebit(base);
	}
	printk("      \b\b\b\b\b\b");	// erase counter...
	return 1;
}

//--ITL08---------------------------------------------------------------------
#define BSIZE 65536

/*
 *  Check to see if there is some flash at the location specified.
 * PART is an Intel 28F008 (4 physical chips with 8bit data path = 8 x 4)
 */
static int it08detect(u32 base) 
{
    u32 id ;

    /* read the manufacturer's id */
    W32(base, 0,0xffffffff);        
    W32(base, 0,0x90909090);        

    /* correct manufacturer's id? (Intel 28F008 in an x4 array) */
    if (R32(base, 0) != 0x89898989)
	id = 0;
    else
	id = R32(base, 4);

    /* correct device id? (Intel 28F008 in an x4 array) */
    if (id < 0xA1A1A1A1)
	id = 0;

    /* Make sure the flashPtr is back in normal mode so 
       that we can do user I/O via firmware in flash */
    W32(base, 0,0xffffffff);
    return id ? DEV_NOLOCK : DEV_ERROR;
}

static int it08write(u32 base, volatile u8 *data, u32 size) 
{ 
	u32 status, i;

	// Erase blocks before writing them again
	for(i=0; i < size; i+=BSIZE)
	{
		/* write delete-block command followed by confirm */
		W32(base, i, 0x20202020);
		W32(base, i, 0xD0D0D0D0);

		/* check the status */
		status = 0;
		while ((status & 0x80808080) != 0x80808080)
			status = R32(base, i);

		  /* Make sure the flash is back in normal mode so 
		     that we can do user I/O via firmware in flash */
		W32(base, 0, 0xFFFFFFFF);
	}

        // - Flash programming - 32 bits wide (EBSA285)
        for(i=0; i<size; i+=4, data+=4)
        {
		// this code will notice that the offset into the rom 
		// needs to swapped...
		if (i <= 0x1F)
			i += 0x20;
		else if ((i >= 0x20) && (i < 0x40))
			i -= 0x20;

		/* set the device into write mode */
		W32(base, i&0xFFFFFFFC, 0x40404040);
		/* write the data */
		W32(base, i&0xFFFFFFFC, *(u32 *)data);

		/* check the status */
		while((status & 0x80808080) != 0x80808080)
			status = R32(base, 0);

		/* Make sure the flash is back in normal mode so 
		that we can do user I/O via firmware in flash */
		W32(base, 0, 0xFFFFFFFF);

//		/* a write error? */
//		return ((status & 0x10101010) == 0x10101010) ? 0 : 1;
        }
}

//--=====---------------------------------------------------------------------
#define KB(N)	(1024*(N))
#define MB(N)	(1024*KB(N))

static const flash_t dev[] =
	{
	//---flash parts for CHALICE/CATS
	{ "SST   29LE020", KB(256),      0, 128,
			sst20detect,     NULL, sst20write },
	{ "Atmel 29LV020", KB(256), KB( 8), 256,
			 at20detect, at20lock,  at20write },
	{ "Atmel 29LV040", KB(512), KB(16),   1,
			 at40detect, at40lock,  at40write },
	//---flash parts for EBSA285
	{ "Intel 28iF008", MB(  4),      0,   4,
			 it08detect,     NULL,  it08write },
	{ NULL }
	};

INLINE int  BKMsize(u32 amount)
{
	if(amount > 1024*1024)
		return amount/(1024*1024);
	if(amount > 1024)
		return amount/(1024);
	return amount;
}

INLINE char BKMunit(u32 amount)
{
	if(amount > 1024*1024)
		return 'M';
	if(amount > 1024)
		return 'K';
	return 'B';
}

static int do_detect(u32 base, char *s)
{
	register int i;

	device = NULL;
	for(i = 0; dev[i].name; i++)
	{
		int r = dev[i].detect(base);

		if(r!=DEV_ERROR)
		{
			device = dev+i;
			sprintf(s, "%s [%4i%c/%4i%c] (%s)",
					dev[i].name,
					BKMsize(dev[i].total_size),
					BKMunit(dev[i].total_size),
					BKMsize(dev[i].block_size),
					BKMunit(dev[i].block_size),
					r==DEV_NOLOCK ? "unlock" : "locked");
			break;
		}
	}
	return device ? 1 : 0;
}

void rom_check(u32 baseaddr)
{
	char s[256];

	printk("Scanning...");
	if(!do_detect(baseaddr, s))
		printk("unknown flash device!!\n");
	else
		printk("%s\n", s);
}

void rom_lock(u32 baseaddr)
{
	printk("Locking...");
	if(!device)
		printk("unknown flash device!!\n");
	else if(!device->lock(baseaddr))
		printk("error!\n");
	else
		printk("done!\n");
}

void rom_burn(u32 baseaddr, u8 *data, u32 size)
{
	printk("Burning...");
	if(!device)
		printk("unknown flash device!!\n");
	else if(!device->write(baseaddr, data, size))
		printk("error!\n");
	else
		printk("done!\n");
}

