/* MKrom
 *
 * David Aparicio                   daparic@terra.es - Mar 2001
 *
 * Original idea from "mkaif.c", by Russell King
 * Adapted (hacked ?) for ELF by Benjamin Herrenschmidt
 *
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <endian.h>
#include <elf.h>

#define __USE_GLIBC__
#include "../libBIOS/sys_crc32.c"

#if BYTE_ORDER == BIG_ENDIAN
#define SWAP16(x)	bswap_16(x)
#define SWAP32(x)	bswap_32(x)
#else
#define SWAP16(x)	(x)
#define SWAP32(x)	(x)
#endif

typedef struct {
	u8 *code;	// malloc chunk for CODE section of this image
	u8 *data;	// malloc chunk for DATA section of this image
	enum { ELF_T=1, BIN_T=2 }   type;	// ELF/ROM
	mbE_t *dir;	// descriptor for this image
	} table_t;

#define FL_ELF(TB)  ((TB)==ELF_T)
#define FL_BIN(TB)  ((TB)==BIN_T)

#define IS_ELF(T)  FL_ELF(table[T].type)
#define IS_BIN(T)  FL_BIN(table[T].type)
#define IS_POST(T) IMG_POST((mbA_t *)table[T].code)
#define IS_LOAD(T) IMG_LOAD((mbA_t *)table[T].code)
#define IS_OTHR(T) IMG_OTHR((mbA_t *)table[T].code)

static table_t table[MAX_FILES];

#define MAX_PHDRS	10

static void show_dir(mbE_t *d, u32 type, int tab)
{
	char buffer[256];

	memset(buffer, ' ', sizeof(buffer));
	printf("%.*s %08X(%6i) %08X(%6i) %s/%s/%s - %s\n",
		tab*3, buffer,
		d->baseC, d->sizeC,
		d->baseD, d->sizeD,
		FL_BIN(type) ? "BIN" : "ELF",
		d->type==SFS_TYPE_POST
			? "POST"
			: d->type==SFS_TYPE_LOAD
				? "LOAD"
				: d->type==SFS_TYPE_BZIP
					? "BZIP"
					: "EXEC",
		d->arch==SFS_ARCH_SA110
			? "ARM"
			: d->arch==SFS_ARCH_DSP
				? "DSP"
				: "LCA",
		d->label);
	if((u32)d->entry != (u32)d->baseC)
		printf("entry at @%08X\n", d->entry);
}

static void show(u32 i, int tab)
{
	printf("CRC32(%08X)",
		crc32(table[i].code, table[i].dir->sizeC));
	show_dir(table[i].dir, table[i].type, tab);
	if(IS_LOAD(i) && IS_BIN(i))
	{
		u32 j;
		mbA_t *mbA = (mbA_t *)table[i].code;

		for(j=1; (*mbA)[j].offset; j++)
		{
			printf("CRC32(%08X)",
				crc32(table[i].code + (*mbA)[j].offset,
					(*mbA)[j].sizeC));
			show_dir((*mbA)+j, BIN_T, tab+1);
		}
	}
}

static int headerROM(int infile, u32 *posC, u32 *posD, mbE_t *dir)
{
	mbA_t hdr;
	u32   len;

	len = lseek(infile, 0, SEEK_END);
	if(len < 0) {
		fprintf(stderr, "lseek: %s\n", strerror(errno));
		return 0;
	}
	if (lseek(infile, 0, SEEK_SET) < 0) {
		fprintf(stderr, "lseek: %s\n", strerror(errno));
		return 0;
	}
	if(read(infile, &hdr, sizeof(hdr)) != sizeof(hdr))
	{
		fprintf(stderr, "ROM.readHdr: %s\n", strerror(errno));
		return 0;
	}

#ifdef DEPURACION
fprintf(stderr, "BIN file detected\n");
#endif
	*posC = 0;
	dir->baseC = hdr[0].baseC;
	dir->baseD = hdr[0].baseD;
	dir->sizeC = hdr[0].sizeC;
	dir->sizeD = hdr[0].sizeD;
	dir->type  = hdr[0].type;
	dir->arch  = hdr[0].arch;
	dir->entry = hdr[0].entry;
	dir->offset= hdr[0].offset;

	if(IMG_POST(&hdr))
	{
		*posD      = dir->sizeC;
		return BIN_T;
	}
	if(IMG_LOAD(&hdr))
	{
		*posD      = len - dir->sizeD;
		return BIN_T;
	}
	if(hdr[0].arch == SFS_ARCH_LCA)
	{
		*posC     += sizeof(mbE_t);
		dir->baseC = hdr[0].baseC;
		dir->baseD = 0;
		dir->sizeC = hdr[0].sizeC - sizeof(mbE_t);
		dir->sizeD = 0;
		dir->type  = hdr[0].type;
		dir->arch  = hdr[0].arch;
		dir->entry = 0;
		dir->offset= 0;
		*posD      = len - dir->sizeD;
		return BIN_T;	// A standalone binary (internal image)
	}
	return 0;
}

static int headerELF(int infile, u32 *posC, u32 *posD, mbE_t *dir)
{
	int bytes, i;
	mbE_t mbE;
	int data_phead, text_phead;
	Elf32_Ehdr elf_head;
	Elf32_Phdr p_head[MAX_PHDRS];
	Elf32_Shdr s_head[MAX_PHDRS];

	if (lseek(infile, 0, SEEK_SET) < 0) {
		fprintf(stderr, "lseek: %s\n", strerror(errno));
		return 0;
	}

	bytes = read(infile, &elf_head, sizeof(elf_head));
	if (bytes < 0) {
		fprintf(stderr, "read: %s\n", strerror(errno));
		return 0;
	} else if (bytes != sizeof(elf_head)) {
		fprintf(stderr, "file truncated\n");
		return 0;
	}

	if	(  (elf_head.e_ident[EI_MAG0] != ELFMAG0)
		|| (elf_head.e_ident[EI_MAG1] != ELFMAG1)
		|| (elf_head.e_ident[EI_MAG2] != ELFMAG2)
		|| (elf_head.e_ident[EI_MAG3] != ELFMAG3))
			return 0;
	data_phead = text_phead = -1;

	if(SWAP32(elf_head.e_phoff))
	{
 		lseek(infile, SWAP32(elf_head.e_phoff), SEEK_SET);
		bytes = read(infile, &p_head,
			SWAP16(elf_head.e_phnum) * sizeof(Elf32_Phdr));
	}
	if(SWAP32(elf_head.e_shoff))
	{
 		lseek(infile, SWAP32(elf_head.e_shoff), SEEK_SET);
		bytes = read(infile, &s_head,
			SWAP16(elf_head.e_shnum) * sizeof(Elf32_Shdr));
	}

	for (i = 0; i < SWAP16(elf_head.e_shnum); ++i)
	{
		if (SWAP32(s_head[i].sh_type) == SHT_PROGBITS)
		{
			if (SWAP32(s_head[i].sh_flags) & SHF_EXECINSTR)
			{
#ifdef DEPURACION
				printf("ELF segment %i %08X(%i) offs %i\n",
					i,
					SWAP32(s_head[i].sh_addr),
					SWAP32(s_head[i].sh_size),
					SWAP32(s_head[i].sh_offset));
#endif
				dir->entry = SWAP32(elf_head.e_entry);
				dir->baseC = (void *)SWAP32(s_head[i].sh_addr);
				dir->sizeC = SWAP32(s_head[i].sh_size);
				*posC      = SWAP32(s_head[i].sh_offset);
				dir->sizeC = ((dir->sizeC+3)/4)*4;
			}
			if (SWAP32(s_head[i].sh_flags) & SHF_WRITE)
			{
#ifdef DEPURACION
				printf("ELF segment %i %08X(%i) offs %i\n",
					i,
					SWAP32(s_head[i].sh_addr),
					SWAP32(s_head[i].sh_size),
					SWAP32(s_head[i].sh_offset));
#endif
				dir->baseD = (void *)SWAP32(s_head[i].sh_addr);
				dir->sizeD = SWAP32(s_head[i].sh_size);
				*posD      = SWAP32(s_head[i].sh_offset);
				dir->sizeD = ((dir->sizeD+3)/4)*4;
			}
		}
	}

	for (i = 0; i < SWAP16(elf_head.e_phnum); ++i)
	{
#ifdef DEPURACION
		printf("ELF section %i p_type=%04X, p_flags=%04X\n",
				i,
				SWAP32(p_head[i].p_type),
				SWAP32(p_head[i].p_flags));
#endif
		if (SWAP32(p_head[i].p_type) == PT_LOAD)
		{
			if (SWAP32(p_head[i].p_flags) & PF_X)
			{
			  if(text_phead == -1)
			  {
				dir->entry = SWAP32(elf_head.e_entry);
				dir->baseC = (void *)SWAP32(p_head[i].p_vaddr);
				dir->sizeC = SWAP32(p_head[i].p_filesz);
				*posC      = SWAP32(p_head[i].p_offset);
				text_phead = i;
				dir->sizeC = ((dir->sizeC+3)/4)*4;
			  }
			}
			else
			{
			  if(data_phead == -1)
			  {
				dir->baseD = (void *)SWAP32(p_head[i].p_vaddr);
				dir->sizeD = SWAP32(p_head[i].p_filesz);
				*posD      = SWAP32(p_head[i].p_offset);
				data_phead = i;
				dir->sizeD = ((dir->sizeD+3)/4)*4;
			  }
			}
		}
	}

	// read "magic signature" from CODE segment (POST, LOADER or BINARY)
	lseek(infile, *posC, SEEK_SET);
	read(infile, &mbE, sizeof(mbE));

	dir->offset = mbE.offset;
	dir->arch   = elf_head.e_machine;
	dir->type   = mbE.type==SFS_TYPE_POST || mbE.type==SFS_TYPE_LOAD
			? mbE.type
			: SFS_TYPE_EXEC;
#ifdef DEPURACION
fprintf(stderr, "ELF file detected\n");
#endif
	if(dir->type==SFS_TYPE_POST)
	{
		if(dir->sizeD)
		{
			fprintf(stderr, "POST con segmento de DATOS?\n");
			return 0;
		}
		return ELF_T;
	}
	return ELF_T;
}
	  	
static int add(int to, char *name)
{
	int inF;
	u32 posC, posD, type;
	table_t *t = table + to;

	t->dir = malloc(sizeof(mbE_t));
	memset (t->dir, 0, sizeof(mbE_t));
	if((inF = open(name, O_RDONLY)) < 0)
	{
		free(t->dir);
		return 0;
	}
	name   = strchr(name, '/') ? strrchr(name, '/')+1 : name;
	strncpy(t->dir->label, name, sizeof(t->dir->label)-1);

	t->type = headerELF(inF, &posC, &posD, t->dir);
	if(!t->type && !(t->type = headerROM(inF, &posC, &posD, t->dir)))
	{
		free(t->dir);
		fprintf(stderr, "%s: Not recognized (ELF/ROM)\n", name);
		return 0;
	}
#ifdef DEPURACION
printf(" CODE %08X offs %i size %i, DATA %08X offs %i size %i [typ=%i]\n",
		t->dir->baseC, posC, t->dir->sizeC,
		t->dir->baseD, posD, t->dir->sizeD,
		t->type);
#endif
	if(t->dir->sizeC)
	{
		t->code = malloc(t->dir->sizeC);
		//-- load CODE segment
		lseek(inF, posC, SEEK_SET);
		memset(t->code, 0, t->dir->sizeC);
		read(inF, t->code, t->dir->sizeC);
#ifdef DEPURACION
		fprintf(stderr, "  [code] %02X %02X %02X %02X\n",
			t->code[0], t->code[1], t->code[2], t->code[3]);
#endif
		if(t->dir->type==SFS_TYPE_LOAD || t->dir->type==SFS_TYPE_POST)
		{
#ifdef DEPURACION
			u32 *a = (u32 *)t->code;
			u32 *b = (u32 *)t->dir;
fprintf(stderr, "copy_dir\n"
		"  [code] %08X %08X %08X %08X\n"
		"  [pdir] %08X %08X %08X %08X\n",
			a[0], a[1], a[2], a[3],
			b[0], b[1], b[2], b[3]);
#endif
			memcpy(t->code, t->dir, sizeof(mbE_t));
			free(t->dir);
			t->dir = (mbE_t *)t->code;
		}
#ifdef DEPURACION
		else
		{
			u32 *a = (u32 *)t->code;
			u32 *b = (u32 *)t->dir;
fprintf(stderr, "exec_dir (no copy)\n"
		"  [code] %08X %08X %08X %08X\n"
		"  [pdir] %08X %08X %08X %08X\n",
			a[0], a[1], a[2], a[3],
			b[0], b[1], b[2], b[3]);
		}
#endif
	}
	else
		t->code = NULL;

	if(t->dir->sizeD)
	{
		t->data = malloc(t->dir->sizeD);
		//-- load DATA segment
		lseek(inF, posD, SEEK_SET);
		memset(t->data, 0, t->dir->sizeD);
		read(inF, t->data, t->dir->sizeD);
	}
	else
		t->data = NULL;

	close(inF);
	return 1;
}

void usage(char *s)
{
	fprintf(stderr,
	    "Usage: %s <image.rom> [args]\n"
	    "  where 'args' can be:\n"
	    "   <step1st.elf> <blocksize_in_Kbytes>\n"
	    "   <step2nd.elf> [<step3rd.rom> | <step3rd.elf> ...]\n", s);
	exit(1);
}

int main(int argc, char *argv[])
{
	int idx;
	u32 total;

	if (argc == 1 || argc == 3)
		usage(argv[0]);
	else if(argc == 2)
	{
		if(add(0, argv[1]))
			show(0, 0);
		else
			usage(argv[0]);
		return 0;
	}

	if(!add(0, argv[2]))
	{
		fprintf(stderr, "%s: %s\n", argv[1], strerror(errno));
		return 1;
	}
	if(!IS_ELF(0))
		usage(argv[0]);

	if(IS_POST(0))
	{
		u32 arg = atoi(argv[3]);
		printf("  BOOTSECTOR IS %i Kbytes\n", arg);
		table[0].dir->baseD = table[0].dir->baseC + arg*1024;
		show(0, 0);
	}
	else if(IS_LOAD(0))
	{
		u32 suma;
		u8    *ptr;
		mbA_t *mbA;

		show(0, 0);
		for(idx=1; idx < argc-2; idx++)
		{
			if(!add(idx, argv[idx+2]))
			{
				fprintf(stderr, "%s: %s\n",
					argv[idx+2], strerror(errno));
				return 1;
			}
			show(idx, 0);
		}

		suma = table[0].dir->sizeC;
		for(idx=1; idx<argc-2; idx++)
		{
			// 16 byte-alignment
			suma = ((suma+15)/16)*16;
			table[idx].dir->offset = suma;
			suma += table[idx].dir->sizeC;
			suma += table[idx].dir->sizeD;
		}

		ptr = malloc(suma);
		memcpy(ptr, table[0].code, table[0].dir->sizeC);
		free(table[0].code);

		table[0].code = (void *)ptr;
		table[0].dir  = (void *)ptr;
		mbA           = (void *)ptr;
		for(idx=1; idx<argc-2; idx++)
		{
			memcpy((*mbA)+idx,
				table[idx].dir,
				sizeof(mbE_t));
			memcpy(ptr +table[idx].dir->offset,
				table[idx].code,
				table[idx].dir->sizeC);
			memcpy(ptr
				+ table[idx].dir->offset
				+ table[idx].dir->sizeC,
				table[idx].data,
				table[idx].dir->sizeD);
		}
		(*mbA)[argc-2].offset = 0;
		(*mbA)[0].sizeC = suma;

		table[0].type = BIN_T;
		memset(table[0].dir->label, 0, LEN_LABEL);
		strncpy(table[0].dir->label, argv[1], LEN_LABEL-1);
		printf("                ========         ========\n");
		show(0, 0);
	}
	else
		usage(argv[0]);


//---ARG1: Output file
	{
		mbE_t *mbE = (mbE_t *)table[0].code;
		int outF = open(argv[1], O_RDWR |O_CREAT |O_TRUNC, 0666);

		if (outF < 0) {
			fprintf(stderr, "%s: %s\n", argv[1], strerror(errno));
			return 1;
		}
		if(mbE->sizeC)
			write(outF, table[0].code, mbE->sizeC);
		if(mbE->sizeD)
			write(outF, table[0].data, mbE->sizeD);
		close(outF);
	}
	return 0;
}
